"""
Internal classes supporting the runtime.
"""
import collections
from collections import Mapping, MutableSequence
import copy
import os
import os.path
import six
from enum import IntEnum

from .eval import RefContext, setEvalFunc, Ref, mapValue
from .result import ResultsMap, Result, ExternalValue, serializeValue
from .util import (intersectDict, mergeDicts, ChainMap, findSchemaErrors,
                UnfurlError, UnfurlValidationError, assertForm, sensitive_str)
from ansible.template import Templar
from ansible.parsing.dataloader import DataLoader

import logging
logger = logging.getLogger('unfurl')

# XXX3 doc: notpresent is a positive assertion of non-existence while notapplied just indicates non-liveness
# notapplied is therefore the default initial state
Status = IntEnum("Status", "ok degraded stopped error pending notapplied notpresent", module=__name__)

# ignore may must
Priority = IntEnum("Priority", "ignore optional required", module=__name__)

# omit discover exist
Action = IntEnum("Action", "discover instantiate revert", module=__name__)

class Defaults(object):
  shouldRun = Priority.required
  intent = Action.instantiate

class File(ExternalValue):
  """
  Represents a local file.
  get() returns the given file path (usually relative)
  """
  def __init__(self, name, baseDir=''):
    super(File, self).__init__('file', name)
    self.baseDir = baseDir or ''

  def getFullPath(self):
    return os.path.abspath(os.path.join(self.baseDir, self.get()))

  def resolveKey(self, name=None, currentResource=None):
    """
    path # absolute path
    contents # file contents (None if it doesn't exist)
    """
    if not name:
      return self.get()

    if name == 'path':
      return self.getFullPath()
    elif name == 'contents':
      with open(self.getFullPath(), 'r') as f:
        return f.read()
    else:
      raise KeyError(name)

setEvalFunc('file', lambda arg, ctx: File(mapValue(arg, ctx), ctx.baseDir))

# XXX need an api check if an object was marked sensitive
# _secrets = weakref.WeakValueDictionary()
# def addSecret(secret):
#   _secrets[id(secret.get())] = secret
# def isSecret(obj):
#    return id(obj) in _secrets

class SensitiveValue(ExternalValue):
  def __init__(self, value):
    super(SensitiveValue, self).__init__('sensitive', value)

  def asRef(self, options=None):
    return sensitive_str(self.get())

setEvalFunc('sensitive', lambda arg, ctx: SensitiveValue(mapValue(arg, ctx)))

def isSensitive(obj):
  return isinstance(obj, (sensitive_str, SensitiveValue, SecretResource))

def _getTemplateTestRegEx():
  return Templar(DataLoader())._clean_regex
_clean_regex = _getTemplateTestRegEx()

def isTemplate(val, ctx):
  return isinstance(val, six.string_types) and not not _clean_regex.search(val)

class _VarTrackerDict(dict):
  def __getitem__(self, key):
    val = super(_VarTrackerDict, self).__getitem__(key)
    if isinstance(val, Result):
      self.ctx.referenced.addReference(key, val)
      return val.resolved
    else:
      self.ctx.referenced.addReference(key, Result(val))
      return val

def applyTemplate(value, ctx):
  # implementation notes:
  #   see https://github.com/ansible/ansible/test/units/template/test_templar.py
  #   dataLoader is only used by _lookup and to set _basedir (else ./)
  if ctx.baseDir and ctx.templar._basedir != ctx.baseDir:
    # we need to create a new templar
    loader = DataLoader()
    if ctx.baseDir:
      loader.set_basedir(ctx.baseDir)
    templar = Templar(loader)
    ctx.templar = templar

  vars = _VarTrackerDict(__unfurl = ctx)
  vars.update(ctx.vars)
  vars.ctx = ctx

  # replaces current vars
  # don't use setter to avoid isinstance(dict) check
  ctx.templar._available_variables = vars
  index = ctx.referenced.start()

  oldvalue = value
  # set referenced to track references (set by Ref.resolve)
  # need a way to turn on and off
  try:
    # strip whitespace so jinija native types resolve even with extra whitespace
    # disable caching so we don't need to worry about the value of a cached var changing
    value = ctx.templar.template(value.strip(), cache=False)
    if value != oldvalue:
      ctx.trace("processed template:", value)
      for result in ctx.referenced.getReferencedResults(index):
        if isSensitive(result.external or result.resolved):
          ctx.trace("setting template result as sensitive")
          return SensitiveValue(value) # mark the template result as sensitive
  finally:
    ctx.referenced.stop()
  return value

setEvalFunc('template', applyTemplate)

def runLookup(name, templar, *args, **kw):
  from ansible.plugins.loader import lookup_loader
  # https://docs.ansible.com/ansible/latest/plugins/lookup.html
  # "{{ lookup('url', 'https://toshio.fedorapeople.org/one.txt', validate_certs=True) }}"
  #       would end up calling the lookup plugin named url's run method like this::
  #           run(['https://toshio.fedorapeople.org/one.txt'], variables=available_variables, validate_certs=True)
  instance = lookup_loader.get(name, loader = templar._loader, templar = templar)
  # ansible_search_path = []
  result = instance.run(args, variables=templar._available_variables, **kw)
  # XXX check for wantList
  if not result:
    return None
  if len(result) == 1:
    return result[0]
  else:
    return result

def lookupFunc(arg, ctx):
  """
  Runs an ansible lookup plugin. Usage:

  lookup:
     lookupFunctionName: 'arg' or ['arg1', 'arg2']

  Or if you need to pass keyword arguments to the lookup function:

  lookup:
    - lookupFunctionName: 'arg' or ['arg1', 'arg2']
    - kw1: value
    - kw2: value
  """
  arg = mapValue(arg, ctx)
  if isinstance(arg, Mapping):
    assert len(arg) == 1
    name, args = list(arg.items())[0]
    kw = {}
  else:
    assertForm(arg, list)
    name, args = list(assertForm(arg[0]).items())[0]
    kw = dict(list(assertForm(kw).items())[0] for kw in arg[1:])

  if not isinstance(args, MutableSequence):
    args = [args]
  return runLookup(name, ctx.templar, *args, **kw)

setEvalFunc('lookup', lookupFunc)

def getInput(arg, ctx):
  try:
    return ctx.currentResource.root.findResource('inputs').attributes[arg]
  except KeyError:
    raise UnfurlError("undefined input '%s'" % arg)
setEvalFunc('get_input', getInput, True)

def concat(args, ctx):
  return ''.join([str(a) for a in mapValue(args, ctx)])
setEvalFunc('concat', concat, True)

def token(args, ctx):
  args = mapValue(args, ctx)
  return args[0].split(args[1])[args[2]]
setEvalFunc('token', token, True)

# XXX this doesn't work with node_filters, need an instance to get a specific result
def getToscaProperty(args, ctx):
  from toscaparser.functions import get_function
  tosca_tpl = ctx.currentResource.root.template.toscaEntityTemplate
  node_template = ctx.currentResource.template.toscaEntityTemplate
  return get_function(tosca_tpl, node_template, {'get_property': args}).result()
setEvalFunc('get_property', getToscaProperty, True)

def getImport(arg, ctx):
  """
  Returns the external resource associated with the named import
  """
  try:
    imported = ctx.currentResource.root.imports[arg]
  except KeyError:
    raise UnfurlError("Can't find import '%s'" % arg)
  if arg == 'secret':
    return SecretResource(arg, imported)
  else:
    return ExternalResource(arg, imported)

setEvalFunc('external', getImport)

class ExternalResource(ExternalValue):
  """
  Wraps a foreign resource
  """
  def __init__(self, name, importSpec):
    super(ExternalResource, self).__init__('external', name)
    self.resource = importSpec.resource
    self.schema = importSpec.spec.get('properties')

  def _validate(self, obj, schema, name):
    if schema:
      messages = findSchemaErrors(serializeValue(obj), schema)
      if messages:
        raise UnfurlValidationError("schema validation failed for attribute '%s': %s" % (name, messages[1]), messages[1])

  def _getSchema(self, name):
    return self.schema and self.schema.get(name, {})

  def get(self):
    return self.resource

  def resolveKey(self, name=None, currentResource=None):
    if not name:
      return self.resource

    schema = self._getSchema(name)
    try:
      value = self.resource._resolve(name)
      # value maybe a Result
    except KeyError:
      if schema and 'default' in schema:
        return schema['default']
      raise

    if schema:
      self._validate(value, schema, name)
    # we don't want to return a result across boundaries
    return value

class SecretResource(ExternalResource):
  def resolveKey(self, name=None, currentResource=None):
    # raises KeyError if not found
    return super(SecretResource, self).resolveKey(name, currentResource)

# shortcuts for local and secret
def shortcut(arg, ctx):
  return Ref(dict(ref=dict(external=ctx.currentFunc), foreach=arg)).resolve(ctx, wantList='result')
setEvalFunc('local', shortcut)
setEvalFunc('secret', shortcut)

class DelegateAttributes(object):
  def __init__(self, interface, resource):
    self.interface = interface
    self.resource = resource
    if interface == 'inherit':
      self.inheritFrom = resource.attributes['inheritFrom']
    if interface == 'default':
      self.default = resource.attributes['default']

  def __call__(self, key):
    if self.interface == 'inherit':
      return self.inheritFrom.attributes[key]
    elif self.interface == 'default':
      result = Ref(self.default).resolve(RefContext(self.resource, vars=dict(key=key)))
      if not result:
        raise KeyError(key)
      elif len(result) == 1:
        return result[0]
      else:
        return result

class ResourceChanges(collections.OrderedDict):
  """
  Records changes made by configurations.
  Serialized as the "modifications" properties

  modifications:
    resource1:
      attribute1: newvalue
      attribute2: %delete # if deleted
      .added: # set if resource was added
      .status: # set when status changes, including when removed (Status.notpresent)
  """
  statusIndex = 0
  addedIndex = 1
  attributesIndex = 2

  def sync(self, resource):
    """ Update self to only include changes that are still live"""
    for k, v in list(self.items()):
      current = Ref(k).resolveOne(RefContext(resource))
      if current:
        attributes = v[self.attributesIndex]
        if attributes:
          v[self.attributesIndex] = intersectDict(attributes, current._attributes)
        if v[self.statusIndex] != current._localStatus:
          v[self.statusIndex] = None
      else:
        del self[k]

  def addChanges(self, changes):
    for name, change in changes.items():
      old = self.get(name)
      if old:
        old[self.attributesIndex] = mergeDicts(old[self.attributesIndex], change)
      else:
        self[name] = [None, None, change]

  def addStatuses(self, changes):
    for name, change in changes.items():
      assert not isinstance(change[1], six.string_types)
      old = self.get(name)
      if old:
        old[self.statusIndex] = change[1]
      else:
        self[name] = [change[1], None, {}]

  def addResources(self, resources):
    for resource in resources:
      self['::'+resource['name']] = [None, resource, None]

  def updateChanges(self, changes, statuses, resource):
    self.addChanges(changes)
    self.addStatuses(statuses)
    if resource:
      self.sync(resource)

class AttributeManager(object):
  """
  Tracks changes made to Resources

  Configurator set attributes override spec attributes.
  A configurator can delete an attribute but it will not affect the spec attributes
  so deleting an attribute is essentially restoring the spec's definition of the attribute
  (if it is defined in a spec.)
  Changing an overridden attribute definition in the spec will have no effect
  -- if a configurator wants to re-evaluate that attribute, it can create a dependency on it
  so to treat that as changed configuration.
  """
  # what about an attribute that is added to the spec that already exists in status?
  # XXX2 tests for the above behavior
  def __init__(self):
    self.attributes = {}
    self.statuses = {}

  def setStatus(self, resource, newvalue):
    assert newvalue is None or isinstance(newvalue, Status)
    if resource.key not in self.statuses:
      self.statuses[resource.key] = [resource._localStatus, newvalue]
    else:
      self.statuses[resource.key][1] = newvalue

  def getAttributes(self, resource):
    if resource.key not in self.attributes:
      if resource.template:
        specd = resource.template.properties
        defaultAttributes = resource.template.defaultAttributes
        _attributes = ChainMap(copy.deepcopy(resource._attributes), specd, defaultAttributes)
      else:
        _attributes = ChainMap(copy.deepcopy(resource._attributes))

      attributes = ResultsMap(_attributes, RefContext(resource))
      self.attributes[resource.key] = (resource, attributes)
      return attributes
    else:
      return self.attributes[resource.key][1]

  # def revertChanges(self):
  #   self.attributes = {}
  #   # for resource, old, new in self.statuses.values():
  #   #   resource._localStatus = old

  def commitChanges(self):
    changes = {}
    for resource, attributes in self.attributes.values():
      # save in _attributes in serialized form
      overrides, specd = attributes._attributes.split()
      resource._attributes = {}
      defs = resource.template and resource.template.attributeDefs or {}
      foundSensitive = []
      for key, value in overrides.items():
        if not isinstance(value, Result):
          # hasn't been touched so keep it as is
          resource._attributes[key] = value
        elif key not in specd or value.hasDiff():
          # value is a Result and it is either new or different from the original value, so save
          defMeta = key in defs and defs[key].schema.get('metadata', {}) or {}

          #XXX if defMeta.get('immutable') and key in specd:
          #  error('value of attribute "%s" changed but is marked immutable' % key)

          if defMeta.get('sensitive'):
            # attribute marked as sensitive and value isn't a secret so mark value as sensitive
            if not value.external: # externalvalues are ok since they don't reveal much
              # we won't be able to restore this value since we can't save it
              resource._attributes[key] = sensitive_str(value.resolved)
              foundSensitive.append(key)
              continue
          resource._attributes[key] = value.asRef() #serialize Result

      # save changes
      diff = attributes.getDiff()
      if not diff:
        continue
      for key in foundSensitive:
        if key in diff:
          diff[key] = resource._attributes[key]
      changes[resource.key] = diff

    self.attributes = {}
    # self.statuses = {}
    return changes



"""
Classes for managing the local environment.

Repositories can optionally be organized into projects that have a local configuration.

There is always a home project that contains the local secret instances.
"""
import os
import os.path
# import six
from .repo import Repo
from .util import (UnfurlError)
from .yamlloader import YamlConfig

DefaultManifestName = 'manifest.yaml'
DefaultLocalConfigName = 'unfurl.yaml'
DefaultHomeDirectory = '.unfurl_home'
HiddenMarkerName = '.unfurl'

class Project(object):
  """
  A Unfurl project folder is a folder that contains

  The spec and instance repo and an local configuration file (unfurl.yaml)

  project/spec/.git # has template.yaml and manifest-template.yaml
         instances/current/.git # has manifest.yaml
          # subprojects are created by repository declarations in spec or instance
          subproject/spec/
                    instances/current
          unfurl.yaml # might create 'secret' or 'local' subprojects
          revisions/...
  """
  def __init__(self, path, localEnv):
    if os.path.isdir(path):
      self.projectRoot = path
      test = os.path.join(self.projectRoot, localEnv.DefaultLocalConfigName)
      if os.path.exists(test):
        self.localConfig = LocalConfig(test, localEnv.homeProject and localEnv.homeProject.localConfig)
      elif localEnv.homeProject:
        self.localConfig = localEnv.homeProject.localConfig
      else:
        self.localConfig = LocalConfig()
    else:
      self.projectRoot = os.path.dirname(path)
      self.localConfig = LocalConfig(path, localEnv.homeProject and localEnv.homeProject.localConfig)

    self.workingDirs = Repo.findGitWorkingDirs(self.projectRoot)

  def getRepos(self):
    return [repo for (gitUrl, repo) in self.workingDirs.values()]

  def getCurrentInstanceRepo(self):
    return os.path.join(self.projectRoot, 'instances', 'current')

  def findDefaultInstanceManifest(self):
    fullPath = self.localConfig.getDefaultManifestPath()
    if fullPath:
      if not os.path.exists(fullPath):
        raise UnfurlError("The default manifest found in %s does not exist: %s" % (self.localConfig.config.path, os.path.abspath(fullPath)))
    else:
      fullPath = os.path.join(self.getCurrentInstanceRepo(), DefaultManifestName)
      if not os.path.exists(fullPath):
        raise UnfurlError("The default manifest does not exist: %s" % os.path.abspath(fullPath))
    return fullPath

  def isPathInProject(self, path):
    return os.path.abspath(self.projectRoot)+os.sep in os.path.abspath(path)+os.sep

  def _createPathForGitRepo(self, gitUrl):
    basename = name = os.path.splitext(os.path.basename(gitUrl))[0]
    counter = 1
    while os.path.exists(os.path.join(self.projectRoot, name)):
      name = basename + str(counter)
      counter += 1
    return os.path.join(self.projectRoot, name)

  def findGitRepo(self, repoURL, revision=None):
    candidate = None
    for dir, (url, repo) in self.workingDirs.items():
      if repoURL == url:
        if not revision or revision == repo.revision:
          return repo
        else:
          candidate = repo
    return candidate

  def findPathInRepos(self, path, importLoader=None):
    candidate = None
    for dir, (url, repo) in self.workingDirs.items():
      filePath, revision, bare = repo.findPath(path, importLoader)
      if filePath:
        if not bare:
          return repo, filePath, revision, bare
        else:
          candidate = (repo, filePath, revision, bare)
    return candidate or None, None, None, None

  def createWorkingDir(self, gitUrl, revision='HEAD'):
    localRepoPath = self._createPathForGitRepo(gitUrl)
    repo = Repo.createWorkingDir(gitUrl, localRepoPath, revision)
    # add to workingDirs
    self.workingDirs[localRepoPath] = (gitUrl, repo)
    return repo

class LocalConfig(object):
  """
  The local configuration.

  - a list of top-level projects
  - list of instance manifests with their local configuration
  - the default local and secret instances

unfurl:
  version:

instances:
  - file:
    repository:
    # default instance if there are multiple instances in that project
    # (only applicable when config is local to a project)
    default: true
    local:
      file: path
      repository:
      resource: root
      # or:
      attributes:
        #XXX: inheritFrom:
    secret:

defaults: # used if the manifest isn't defined above
 local:
 secret:

projects:
  - path:
    default: True
    instance: instances/current
    spec: spec
"""
  def __init__(self, path=None, parentConfig=None):
    defaultConfig = {}
    # XXX define schema and validate
    self.config = YamlConfig(defaultConfig, path=path)
    self.manifests = self.config.config.get('instances', [])
    self.defaults = self.config.config.get('defaults', {})
    self.parentConfig = parentConfig

  def adjustPath(self, path):
    """
    Makes sure relative paths are relative to the location of this local config
    """
    return os.path.join(self.config.getBaseDir(), path)

  def getDefaultManifestPath(self):
    if len(self.manifests) == 1:
      return self.adjustPath(self.manifests[0]['file'])
    else:
      for spec in self.manifests:
        if spec.get('default'):
          return self.adjustPath(spec['file'])
    return None

  def getLocalResource(self, manifestPath, localName, importSpec):
    """
    localName is either 'local' or 'secret'
    """
    from .runtime import Resource
    localRepo = None
    for spec in self.manifests:
      if manifestPath == self.adjustPath(spec['file']):
        localRepo = spec.get(localName)
    if not localRepo:
      localRepo = self.defaults.get(localName)
    if not localRepo and self.parentConfig:
      return self.parentConfig.getLocalResource(manifestPath, localName, importSpec)

    if localRepo:
      attributes = localRepo.get('attributes')
      if attributes is not None:
        if 'default' in attributes:
          if not 'default' in attributes.get('.interfaces', {}):
            attributes.setdefault('.interfaces', {})['default'] = 'unfurl.support.DelegateAttributes'
        if 'inheritFrom' in attributes:
          if not 'inherit' in attributes.get('.interfaces', {}):
            attributes.setdefault('.interfaces', {})['inherit'] = 'unfurl.support.DelegateAttributes'
          if attributes['inheritFrom'] == 'home' and self.parentConfig:
            parent = self.parentConfig.getLocalResource(manifestPath, localName, importSpec)
            localResource = Resource(localName, attributes)
            if parent:
              localResource._attributes['inheritFrom'] = parent
              return localResource
            else:
              importSpec['inheritHack'] = localResource
              return None
        repoResource = Resource(localName, attributes)
        repoResource.baseDir = self.config.getBaseDir()
        return repoResource
      else:
        # the local or secret is a resource defined in a local manifest
        # set the url and resource name so the importing manifest loads it
        # XXX but should load here and save for re-use
        importSpec.update(localRepo)
        if 'file' in localRepo:
          importSpec['file'] = self.adjustPath(localRepo['file'])
        if 'repository' in localRepo:
          importSpec['repository'] = self.adjustPath(localRepo['repository'])
        return None

    # none found, return empty resource
    return Resource(localName)

class LocalEnv(object):
  """
  The class represents the local environment that a instance manifest runs in.

  The instance manifest and/or the current project
  The local configuration
  """
  homeProject = None

  def __init__(self, manifestPath=None, homepath=None):
    """
    If manifestPath is None find the first .unfurl or manifest.yaml
    starting from the current directory.
    """
    # XXX need to save local config when changed
    self.homeConfigPath = self.getHomeConfigPath(homepath)
    self.homeProject = Project(self.homeConfigPath, self)

    if manifestPath:
      pathORproject = self.findManifestPath(manifestPath)
    else:
      # not specified: search current directory and parents for either a manifest or a project
      pathORproject = self.searchForManifestOrProject('.')

    if isinstance(pathORproject, Project):
      self.project = pathORproject
      self.manifestPath = pathORproject.findDefaultInstanceManifest()
    else:
      self.manifestPath = pathORproject
      self.project = self.findProject(os.path.dirname(pathORproject))

    self.instanceRepo = self._getInstanceRepo()
    self.config = self.project and self.project.localConfig or self.homeProject.localConfig

  # manifestPath specified
  #  doesn't exist: error
  #  is a directory: either instance repo or a project
  def findManifestPath(self, manifestPath):
    if not os.path.exists(manifestPath):
      raise UnfurlError("Manifest file does not exist: '%s'" % os.path.abspath(manifestPath))

    if os.path.isdir(manifestPath):
      test = os.path.join(manifestPath, DefaultManifestName)
      if os.path.exists(test):
        return test
      else:
        test = os.path.join(manifestPath, DefaultLocalConfigName)
        if os.path.exists(test):
          return Project(test, self)
        else:
          message = "Can't find a unfurl manifest or project in folder '%s'" % manifestPath
          raise UnfurlError(message)
    else:
      return manifestPath

  def _getInstanceRepo(self):
    instanceDir = os.path.dirname(self.manifestPath)
    if self.project and instanceDir in self.project.workingDirs:
      return self.project.workingDirs[instanceDir][1]
    else:
      return Repo.createGitRepoIfExists(instanceDir)

  def getRepos(self):
    if self.project:
      return self.project.getRepos()
    else:
      return [self.instanceRepo]

  def searchForManifestOrProject(self, dir):
    current = os.path.abspath(dir)
    while current and current != os.sep:
      test = os.path.join(current, DefaultManifestName)
      if os.path.exists(test):
        return test

      test = os.path.join(current, DefaultLocalConfigName)
      if os.path.exists(test):
        return Project(test, self)

      current = os.path.dirname(current)

    message = "Can't find a unfurl repository in current directory (or any of the parent directories)"
    raise UnfurlError(message)

  def findProject(self, testPath):
    """
    Walk parents looking for unfurl.yaml
    """
    current = os.path.abspath(testPath)
    while current and current != os.sep:
      test = os.path.join(current, DefaultLocalConfigName)
      if os.path.exists(test):
        return Project(test, self)
      current = os.path.dirname(current)
    return None

  def findMarker(self, testPath):
    test = os.path.join(testPath, HiddenMarkerName)
    if os.path.exists(test):
      return YamlConfig(test)
    else:
      return None

  def getHomeConfigPath(self, homepath):
    if homepath:
      if os.path.isdir(homepath):
        return os.path.abspath(os.path.join(homepath, DefaultLocalConfigName))
      else:
        return os.path.abspath(homepath)
    return os.path.expanduser(os.path.join('~', DefaultHomeDirectory, DefaultLocalConfigName))

  def getLocalResource(self, name, importSpec):
    if name != 'local' and name != 'secret':
      return None
    return self.config.getLocalResource(self.manifestPath, name, importSpec)

  def findGitRepo(self, repoURL, isFile=True, revision=None):
    if self.project:
      repo = self.project.findGitRepo(repoURL, revision)
      if not repo:
        return self.homeProject.findGitRepo(repoURL, revision)

  def findOrCreateWorkingDir(self, repoURL, isFile=True, revision=None, basepath=None):
    repo = self.findGitRepo(repoURL, revision)
    if not repo:
      if self.project and (basepath is None or self.project.isPathInProject(basepath)):
        project = self.project
      else:
        project = self.homeProject
      repo = project.createWorkingDir(repoURL, revision, basepath)
    return repo, repo.workingDir, repo.revision, revision and repo.revision != revision

  def findPathInRepos(self, path, importLoader=None):
    candidate = None
    if self.project:
      repo, filePath, revision, bare = self.project.findPathInRepos(path, importLoader)
      if repo:
        if not bare:
          return repo, filePath, revision, bare
        else:
          candidate = (repo, filePath, revision, bare)

    repo, filePath, revision, bare = self.homeProject.findPathInRepos(path, importLoader)
    if repo:
      if bare and candidate:
        return candidate
      else:
        return repo, filePath, revision, bare
    return None, None, None, None



import sys
import optparse
import six
import traceback
import itertools
from collections import Mapping, MutableSequence
import os.path
from jsonschema import Draft4Validator, validators
from ruamel.yaml.comments import CommentedMap
from ruamel.yaml.scalarstring import ScalarString, FoldedScalarString
import logging
logger = logging.getLogger('unfurl')

 #import pickle
pickleVersion = 2 #pickle.DEFAULT_PROTOCOL

from ansible.plugins.loader import lookup_loader, filter_loader, strategy_loader
lookup_loader.add_directory(os.path.abspath(os.path.dirname(__file__)), True)
filter_loader.add_directory(os.path.abspath(os.path.dirname(__file__)), True)
strategy_loader.add_directory(os.path.abspath(os.path.join(os.path.dirname(__file__),
                      'vendor', 'ansible_mitogen', 'plugins', 'strategy')), False)

class AnsibleDummyCli(object):
  def __init__(self):
    self.options = optparse.Values()
ansibleDummyCli = AnsibleDummyCli()
from ansible.utils import display
display.logger = logging.getLogger('ansible')
class AnsibleDisplay(display.Display):
  def display(self, msg, color=None, stderr=False, screen_only=False, log_only=False):
    if screen_only:
      return
    log_only = True
    return super(AnsibleDisplay, self).display(msg, color, stderr, screen_only, log_only)
import ansible.constants as C
if 'ANSIBLE_NOCOWS' not in os.environ:
  C.ANSIBLE_NOCOWS=1
if 'ANSIBLE_JINJA2_NATIVE' not in os.environ:
  C.DEFAULT_JINJA2_NATIVE=1
ansibleDisplay = AnsibleDisplay()

def initializeAnsible():
  main = sys.modules.get('__main__')
  # XXX make sure ansible.executor.playbook_executor hasn't been loaded already
  main.display = ansibleDisplay
  main.cli = ansibleDummyCli
initializeAnsible()

VERSION = 'unfurl/v1alpha1' # api version

class UnfurlError(Exception):
  def __init__(self, message, saveStack=False, log=False):
    if saveStack:
      (type, value, traceback) = sys.exc_info()
      if value:
        message += ': ' + str(value)
    super(UnfurlError, self).__init__(message)
    self.stackInfo =  (type, value, traceback) if saveStack and value else None
    if log:
      logger.error(message, exc_info=True)

  def getStackTrace(self):
    if not self.stackInfo:
      return ''
    return ''.join(traceback.format_exception(*self.stackInfo))

class UnfurlValidationError(UnfurlError):
  def __init__(self, message, errors=None):
    super(UnfurlValidationError, self).__init__(message)
    self.errors = errors or []

class UnfurlTaskError(UnfurlError):
  def __init__(self, task, message, log=False):
    super(UnfurlTaskError, self).__init__(message, True, log)
    self.task = task
    task.errors.append(self)

class UnfurlAddingResourceError(UnfurlTaskError):
  def __init__(self, task, resourceSpec):
    resourcename = isinstance(resourceSpec, Mapping) and resourceSpec.get('name', '')
    message = "error creating resource %s" % resourcename
    super(UnfurlTaskError, self).__init__(task, message)
    self.resourceSpec = resourceSpec

class sensitive_str(str):
  pass

def toYamlText(val):
  if isinstance(val, (ScalarString, sensitive_str)):
    return val
  # convert or copy string (copy to deal with things like AnsibleUnsafeText)
  val = str(val)
  if '\n' in val:
    return FoldedScalarString(val)
  return val

def assertForm(src, types=Mapping):
  if not isinstance(src, types):
    raise UnfurlError('Malformed definition: %s' % src)
  return src

# map< apiversion, map<kind, ctor> >
_ClassRegistry = {}
# only one class can be associated with an api interface
def registerClass(apiVersion, kind, factory, replace=False):
  api = _ClassRegistry.setdefault(apiVersion, {})
  if not replace and kind in api:
    if api[kind] is not factory:
      raise UnfurlError('class already registered for %s.%s' % (apiVersion, kind))
  api[kind] = factory

class AutoRegisterClass(type):
  def __new__(mcls, name, bases, dct):
    cls = type.__new__(mcls, name, bases, dct)
    registerClass(VERSION, name, cls)
    return cls

def loadClass(klass, defaultModule='__main__'):
  import importlib
  prefix, sep, suffix = klass.rpartition('.')
  module = importlib.import_module(prefix or defaultModule)
  return getattr(module, suffix, None)

def lookupClass(kind, apiVersion=None, default=None):
  version = apiVersion or VERSION
  api = _ClassRegistry.get(version)
  if api:
    klass = api.get(kind, default)
  else:
    klass = default

  if not klass:
    try:
      klass = loadClass(kind)
    except ImportError:
      klass = None

    if klass:
      registerClass(version, kind, klass, True)
    else:
      raise UnfurlError('Can not find class %s.%s' % (version, kind))
  return klass

def toEnum(enum, value, default=None):
  #from string: Status[name]; to string: status.name
  if isinstance(value, six.string_types):
    return enum[value]
  elif default is not None and not value:
    return default
  else:
    return value

def makeMapWithBase(baseDir):
  def factory(*args, **kws):
    map = CommentedMap(*args, **kws)
    map.baseDir = baseDir
    map.mapCtor = makeMapWithBase(baseDir)
    return map

  return factory

# XXX?? because json keys are strings allow number keys to merge with lists
# values: merge, replace, delete, renamekey
mergeStrategyKey = '+%'
#
def mergeDicts(b, a, cls=dict):
  """
  Returns a new dict (or cls) that recursively merges b into a.
  b is base, a overrides.

  A superset of JSON merge patch (https://tools.ietf.org/html/rfc7386)
  """
  cp = cls()
  skip = []
  for key, val in a.items():
    if key == mergeStrategyKey:
      continue
    if isinstance(val, Mapping):
      strategy = val.get(mergeStrategyKey)
      if key in b:
        bval = b[key]
        if isinstance(bval, Mapping):
          if not strategy:
            strategy = bval.get(mergeStrategyKey, 'merge')
            cls = getattr(bval, 'mapCtor', cls)
            cp[key] = mergeDicts(bval, val, cls)
            continue
          if strategy == 'error':
            raise UnfurlError('merging %s is not allowed, +%: error was set' % key)
      if strategy == 'delete':
        skip.append(key)
        continue
    # XXX merge lists
    # elif isinstance(val, list) and key in b:
    #   bval = b[key]
    #   if isinstance(bval, list):
    #     if appendlists == 'all' or key in appendlists:
    #       cp[key] = bval + val
    #       continue
    #     elif mergelists == 'all' or key in mergelists:
    #       newlist = []
    #       for ai, bi in zip(val, bval):
    #         if isinstance(ai, Mapping) and isinstance(bi, Mapping):
    #           newlist.append(mergeDicts(bi, ai, cls))
    #         elif a1 != deletemarker:
    #           newlist.append(a1)
    #       cp[key] == newlist
    #       continue

    # otherwise a replaces b
    cp[key] = val

  for key, val in b.items():
    if key == mergeStrategyKey:
      continue
    if key not in cp and key not in skip:
      cp[key] = val
  return cp

IncludeKey = '%include'
def getTemplate(doc, key, value, path, cls):
  template = doc
  templatePath = None
  if key == IncludeKey:
    value, template, baseDir = doc.loadTemplate(value)
    cls = makeMapWithBase(baseDir)
  else:
    for segment in key.split('/'):
      # XXX raise error if .. not at start of key
      if segment == '..':
        if templatePath is None:
          templatePath = path[:-1]
        else:
          templatePath = templatePath[:-1]
        template = lookupPath(doc, templatePath, cls)
      # XXX this check should allow array look up:
      if not isinstance(template, Mapping) or segment not in template:
        raise UnfurlError('can not find "%s" in document' % key)
      if templatePath is not None:
        templatePath.append(segment)
      template = template[segment]
    if templatePath is None:
      templatePath = key.split('/')

  try:
    if value != 'raw' and isinstance(template, Mapping): # raw means no further processing
      # if the include path starts with the path to the template
      # throw recursion error
      if key != IncludeKey:
        prefix = list(itertools.takewhile(lambda x: x[0] == x[1], zip(path, templatePath)))
        if len(prefix) == len(templatePath):
          raise UnfurlError('recursive include "%s" in "%s"' % (templatePath, path))
      includes = CommentedMap()
      template = expandDict(doc, path, includes, template, cls=cls)
  finally:
    if key == IncludeKey:
      doc.loadTemplate(baseDir) # pop baseDir
  return template

def hasTemplate(doc, key, path, cls):
  if key == IncludeKey:
    return hasattr(doc, 'loadTemplate')
  template = doc
  for segment in key.split('/'):
    if segment == '..':
      path = path[:-1]
      template = lookupPath(doc, path, cls)
    if not isinstance(template, Mapping):
      raise UnfurlError('included templates changed')
    if segment not in template:
      return False
    template = template[segment]
  return True

class _MissingInclude(object):
  def __init__(self, key, value):
    self.key = key
    self.value = value

  def __repr__(self):
    return self.key

def expandDict(doc, path, includes, current, cls=dict):
  """
  Return a copy of `doc` that expands include directives.
  Include directives look like "+path.to.value"
  When appearing as a key in a map it will merge the result with the current dictionary.
  When appearing as a string or map in a list it will insert the result in the list;
  if result is also a list, each item will be inserted separately.
  (If you don't want that behavior just wrap include in another list, e.g "[+list1]")
  """
  cp = cls()
  # first merge any includes includes into cp
  templates = []
  assert isinstance(current, Mapping), current
  for (key, value) in current.items():
    if key.startswith('+'):
      if key == mergeStrategyKey:
        # cleaner want to skip copying key if not inside a template
        cp[key] = value
        continue
      foundTemplate = hasTemplate(doc, key[1:], path, cls)
      if not foundTemplate:
        includes.setdefault(path, []).append( _MissingInclude(key[1:], value) )
        cp[key] = value
        continue
      includes.setdefault(path, []).append( (key, value) )
      template = getTemplate(doc, key[1:], value, path, cls)
      if isinstance(template, Mapping):
        templates.append( template )
      else:
        if len(current) > 1:
          raise UnfurlError('can not merge non-map value %s' % template)
        else:
          return template # current dict is replaced with a value
    elif key.startswith('q+'):
      cp[key[2:]] = value
    elif isinstance(value, Mapping):
      cp[key] = expandDict(doc, path + (key,), includes, value, cls)
    elif isinstance(value, list):
      cp[key] = list(expandList(doc, path + (key,), includes, value, cls))
    else:
      cp[key] = value

  if templates:
    accum = templates.pop(0)
    templates.append(cp)
    while templates:
      cls = getattr(templates[0], 'mapCtor', cls)
      accum = mergeDicts(accum, templates.pop(0), cls)
    return accum
  else:
    return cp
  # e,g, mergeDicts(mergeDicts(a, b), cp)
  #return includes, reduce(lambda accum, next: mergeDicts(accum, next, cls), templates, {}), cp

def _findMissingIncludes(includes):
  for x in includes.values():
    for i in x:
      if isinstance(i, _MissingInclude):
        yield i

def expandDoc(doc, current=None, cls=dict):
  includes = CommentedMap()
  if current is None:
    current = doc
  if not isinstance(doc, Mapping) or not isinstance(current, Mapping):
    raise UnfurlError('top level element %s is not a dict' % doc)
  expanded = expandDict(doc, (), includes, current, cls)
  last = 0
  while True:
    missing = list(_findMissingIncludes(includes))
    if len(missing) == 0:
      return includes, expanded
    if len(missing) == last: # no progress
      raise UnfurlError('missing includes: %s' % missing)
    last = len(missing)
    includes = CommentedMap()
    expanded = expandDict(expanded, (), includes, current, cls)

def expandList(doc, path, includes, value, cls=dict):
  for i, item in enumerate(value):
    if isinstance(item, six.string_types):
      if item.startswith('+'):
        includes.setdefault(path+(i,), []).append( (item, None) )
        template = getTemplate(doc, item[1:], None, path, cls)
        if isinstance(template, MutableSequence):
          for i in template:
            yield i
        else:
          yield template
      elif item.startswith('q+'):
        yield item[1:]
      else:
        yield item
    elif isinstance(item, Mapping):
      newitem = expandDict(doc, path+(i,), includes, item, cls)
      if isinstance(newitem, MutableSequence):
        for i in newitem:
          yield i
      else:
        yield newitem
    else:
      yield item

def diffDicts(old, new, cls=dict):
  """
  return a dict where old + diff = new
  """
  diff = cls()
  # start with old to preserve original order
  for key, val in old.items():
    if key in new:
      newval = new[key]
      if val != newval:
        if isinstance(val, Mapping) and isinstance(newval, Mapping):
          diff[key] = diffDicts(val, newval, cls)
        else:
          diff[key] = newval
    else:
      diff[key]= {'+%': 'delete'}

  for key in new:
    if key not in old:
      diff[key] = new[key]
  return diff

# XXX rename function, confusing name
def patchDict(old, new, cls=dict):
  """
  Transform old into new while preserving old as much as possible.
  """
  # start with old to preserve original order
  for key, val in list(old.items()):
    if key in new:
      newval = new[key]
      if val != newval:
        if isinstance(val, Mapping) and isinstance(newval, Mapping):
          old[key] = patchDict(val, newval, cls)
        elif isinstance(val, MutableSequence) and isinstance(newval, MutableSequence):
          # preserve old item in list if they are equal to the new item
          old[key] = [(val[val.index(item)] if item in val else item)
                                                    for item in newval]
        else:
          old[key] = newval
    else:
      del old[key]

  for key in new:
    if key not in old:
      old[key] = new[key]

  return old

def intersectDict(old, new, cls=dict):
  """
  remove keys from old that don't match new
  """
  # start with old to preserve original order
  for key, val in list(old.items()):
    if key in new:
      newval = new[key]
      if val != newval:
        if isinstance(val, Mapping) and isinstance(newval, Mapping):
          old[key] = intersectDict(val, newval, cls)
        else:
          del old[key]
    else:
      del old[key]

  return old

def lookupPath(doc, path, cls=dict):
  template = doc
  for segment in path:
    if not isinstance(template, Mapping) or segment not in template:
      return None
    template = template[segment]
  return template

def replacePath(doc, key, value, cls=dict):
  path = key[:-1]
  last = key[-1]
  ref = lookupPath(doc, path, cls)
  ref[last] = value

def addTemplate(changedDoc, path, template):
  current = changedDoc
  key = path.split('/')
  path = key[:-1]
  last = key[-1]
  for segment in path:
    current = current.setdefault(segment, {})
  current[last] = template

def restoreIncludes(includes, originalDoc, changedDoc, cls=dict):
  """
  Modifies changedDoc with to use the includes found in originalDoc
  """
  # if the path to the include still exists
  # resolve the include
  # if the include doesn't exist in the current doc, re-add it
  # create a diff between the current object and the merged includes
  expandedOriginalIncludes, expandedOriginalDoc = expandDoc(originalDoc, cls=cls)
  for key, value in includes.items():
    ref = lookupPath(changedDoc, key, cls)
    if ref is None:
      # inclusion point no longer exists
      continue

    mergedIncludes = {}
    for (includeKey, includeValue) in value:
      if includeKey[1:] == IncludeKey:
        ref = None
        continue
      stillHasTemplate = hasTemplate(changedDoc, includeKey[1:], key, cls)
      if stillHasTemplate:
        template = getTemplate(changedDoc, includeKey[1:], includeValue, key, cls)
      else:
        if hasTemplate(originalDoc, includeKey[1:], key, cls):
          template = getTemplate(originalDoc, includeKey[1:], includeValue, key, cls)
        else:
          template = getTemplate(expandedOriginalDoc, includeKey[1:], includeValue, key, cls)

      if not isinstance(ref, Mapping):
        #XXX3 if isinstance(ref, list) lists not yet implemented
        if ref == template:
          #ref still resolves to the template's value so replace it with the include
          replacePath(changedDoc, key, {includeKey: includeValue}, cls)
        # ref isn't a map anymore so can't include a template
        break

      if not isinstance(template, Mapping):
        # ref no longer includes that template or we don't want to save it
        continue
      else:
        mergedIncludes = mergeDicts(mergedIncludes, template, cls)
        ref[includeKey] = includeValue

      if not stillHasTemplate:
        if includeValue != 'raw':
          if hasTemplate(originalDoc, includeKey[1:], key, cls):
            template = getTemplate(originalDoc, includeKey[1:], 'raw', key, cls)
          else:
            template = getTemplate(expandedOriginalDoc, includeKey[1:], 'raw', key, cls)
        addTemplate(changedDoc, includeKey[1:], template)

    if isinstance(ref, Mapping):
      diff = diffDicts(mergedIncludes, ref, cls)
      replacePath(changedDoc, key, diff, cls)

# https://python-jsonschema.readthedocs.io/en/latest/faq/#why-doesn-t-my-schema-s-default-property-set-the-default-on-my-instance
def extend_with_default(validator_class):
  """
  # Example usage:
  obj = {}
  schema = {'properties': {'foo': {'default': 'bar'}}}
  # Note jsonschema.validate(obj, schema, cls=DefaultValidatingDraft7Validator)
  # will not work because the metaschema contains `default` directives.
  DefaultValidatingDraft7Validator(schema).validate(obj)
  assert obj == {'foo': 'bar'}
  """
  validate_properties = validator_class.VALIDATORS["properties"]

  def set_defaults(validator, properties, instance, schema):
    if not validator.is_type(instance, "object"):
      return

    for key, subschema in properties.items():
      if "default" in subschema:
        instance.setdefault(key, subschema["default"])

    for error in validate_properties(
        validator, properties, instance, schema,
    ):
        yield error

  # new validator class
  return validators.extend(
    validator_class, {"properties" : set_defaults},
  )

DefaultValidatingLatestDraftValidator = extend_with_default(Draft4Validator)

def validateSchema(obj, schema):
  return not findSchemaErrors(obj, schema)

def findSchemaErrors(obj, schema):
  # XXX2 have option that includes definitions from manifest's schema
  validator = DefaultValidatingLatestDraftValidator(schema)
  errors = list(validator.iter_errors(obj))
  if not errors:
    return None
  message = '\n'.join(e.message for e in errors)
  return message, errors

#RefResolver.from_schema(schema)

class ChainMap(Mapping):
  """
  Combine multiple mappings for sequential lookup.
  """
  def __init__(self, *maps):
    self._maps = maps

  def split(self):
    return self._maps[0], ChainMap(*self._maps[1:])

  def __getitem__(self, key):
    for mapping in self._maps:
      try:
        return mapping[key]
      except KeyError:
        pass
    raise KeyError(key)

  def __setitem__(self, key, value):
    self._maps[0][key] = value

  def __iter__(self):
    return iter(frozenset(itertools.chain(*self._maps)))

  def __len__(self):
    return len(frozenset(itertools.chain(*self._maps)))

  def __nonzero__(self):
    return all(self._maps)

  def __repr__(self):
    return "ChainMap(%r)" % (self._maps,)



"""
TOSCA implementation

Differences with TOSCA 1.1:

 * Entity type can allow properties that don't need to be declared
 * Added "any" datatype
 * Interface "implementation" values can be a node template name, and the corresponding
   instance will be used to execute the operation.
"""
from .tosca_plugins import TOSCA_VERSION
from .util import UnfurlValidationError
from .eval import Ref
from toscaparser.tosca_template import ToscaTemplate
# from toscaparser.topology_template import TopologyTemplate
from toscaparser.elements.capabilitytype import CapabilityTypeDef
from toscaparser.common.exception import ExceptionCollector, ValidationError
import logging
logger = logging.getLogger('unfurl')

from toscaparser import functions

class RefFunc(functions.Function):
  def result(self):
      return {self.name: self.args}

  def validate(self):
      pass

functions.function_mappings['eval'] = RefFunc
functions.function_mappings['ref'] = RefFunc

toscaIsFunction = functions.is_function
def is_function(function):
  return toscaIsFunction(function) or Ref.isRef(function)
functions.is_function = is_function

def createDefaultTopology():
  tpl = dict(tosca_definitions_version=TOSCA_VERSION, topology_template=dict(
    node_templates = {'_default': {'type': 'tosca.nodes.Root'}},
    relationship_templates= {'_default': {'type': 'tosca.relationships.Root'}}
    ))
  return ToscaTemplate(yaml_dict_tpl=tpl).topology_template

class ToscaSpec(object):
  ConfiguratorType = 'unfurl.nodes.Configurator'

  def __init__(self, toscaDef, inputs=None, instances=None, path=None):
    # for key in tosca_ext_tpl:
    toscaDef.setdefault('topology_template', dict(
                        node_templates = {},
                        relationship_templates= {}
            ))

    if instances:
      self.loadInstances(toscaDef, instances)

    logger.info("Validating TOSCA template at %s", path)
    try:
      # need to set a path for the import loader
      self.template = ToscaTemplate(path=path, parsed_params=inputs, yaml_dict_tpl=toscaDef)
    except ValidationError:
      message = "\n".join(ExceptionCollector.getExceptionsReport(False))
      raise UnfurlValidationError("TOSCA validation failed for %s: \n%s"  % (path, message),
              ExceptionCollector.getExceptions())

    self.nodeTemplates = {}
    self.configurators = {}
    self.relationshipTemplates = {}
    if hasattr(self.template, 'nodetemplates'):
      for template in self.template.nodetemplates:
        nodeTemplate = NodeSpec(template)
        if template.is_derived_from(self.ConfiguratorType):
          # XXX 'configurator-' is a hack
          self.configurators[template.name[len('configurator-'):]
            if template.name.startswith('configurator-') else template.name] = nodeTemplate
        self.nodeTemplates[template.name] = nodeTemplate
    self.topology = TopologySpec(self.template.topology_template, inputs)

  def getTemplate(self, name):
    if name == '#topology':
      return self.topology
    return self.nodeTemplates.get(name, self.relationshipTemplates.get(name))

  def loadInstances(self, toscaDef, tpl):
    """
    Creates node templates for any instances defined in the spec

    .. code-block:: YAML

      spec:
            instances:
              test:
                implementations:
                  create: test
            implementations:
              test:
                className:    TestSubtaskConfigurator
                inputs:
"""
    node_templates = toscaDef.setdefault('topology_template',{}).setdefault("node_templates", {})
    for name, impl in tpl.get('implementations', {}).items():
      node_templates['configurator-'+name] = self.loadImplementation(impl)

    for name, impl in tpl.get('instances', {}).items():
      if name not in node_templates and impl is not None:
        node_templates[name] = self.loadInstance(impl)

  # load implementations
  #  create configurator template for each implementation
  def loadImplementation(self, _impl):
    impl = _impl.copy()
    implementation = impl.pop('className')
    return dict(type=self.ConfiguratorType,
      properties=impl,
      interfaces={
        'Provides':
          {'run': {
                  'implementation': implementation,
                  }
          }
        })

  def loadInstance(self, impl):
    template = {'type': 'tosca.nodes.Root'} #'unfurl.nodes.Default'}
    implementations = impl.get('implementations')
    if implementations:
      template['interfaces'] = {
        'Standard': dict(pair for pair in implementations.items())
      }
    return template

_defaultTopology =  createDefaultTopology()

# represents a node, capability or relationship
class EntitySpec(object):
  def __init__(self, toscaNodeTemplate):
    self.toscaEntityTemplate = toscaNodeTemplate
    self.name = toscaNodeTemplate.name
    self.type = toscaNodeTemplate.type
    # nodes have both properties and attributes
    # as do capability properties and relationships
    # but only property values are declared
    self.properties = {prop.name: prop.value
        for prop in toscaNodeTemplate.get_properties_objects()}
    if toscaNodeTemplate.type_definition:
      attrDefs = toscaNodeTemplate.type_definition.get_attributes_def_objects()
      self.defaultAttributes = {prop.name: prop.default
              for prop in attrDefs if prop.default is not None}
      propDefs = toscaNodeTemplate.type_definition.get_properties_def()
      propDefs.update(toscaNodeTemplate.type_definition.get_attributes_def())
      self.attributeDefs = propDefs
    else:
      self.defaultAttributes = {}
      self.attributeDefs = {}


  def getInterfaces(self):
    return self.toscaEntityTemplate.interfaces

  def isCompatibleTarget(self, targetStr):
    if self.name == targetStr:
      return True
    return self.toscaEntityTemplate.is_derived_from(targetStr)

  def isCompatibleType(self, typeStr):
    return self.toscaEntityTemplate.is_derived_from(typeStr)

  def getUri(self):
    return self.name # XXX

  def __repr__(self):
    return "EntitySpec('%s')" % self.name

class NodeSpec(EntitySpec):
# has attributes: tosca_id, tosca_name, state, (3.4.1 Node States p.61)
  def __init__(self, template=None):
    if not template:
      template = _defaultTopology.nodetemplates[0]
    EntitySpec.__init__(self, template)

class RelationshipSpec(EntitySpec):
# has attributes: tosca_id, tosca_name
  def __init__(self, template=None):
    if not template:
      template = _defaultTopology.relationship_templates[0]
    EntitySpec.__init__(self, template)

class TopologySpec(EntitySpec):
# has attributes: tosca_id, tosca_name, state, (3.4.1 Node States p.61)
  def __init__(self, template=None, inputs=None):
    if not template:
      template = _defaultTopology
    inputs = inputs or {}

    self.toscaEntityTemplate = template
    self.name = '#topology'
    self.inputs = {input.name: inputs.get(input.name, input.default)
                          for input in template.inputs}
    self.outputs = {output.name: output.value for output in template.outputs}
    self.properties = {}
    self.defaultAttributes = {}

# capabilities.Capability isn't an EntityTemplate but duck types with it
class CapabilitySpec(EntitySpec):
  def __init__(self, name=None, nodeTemplate=None):
    if not nodeTemplate:
      self.nodeTemplate = _defaultTopology.nodetemplates[0]
      cap = self.nodeTemplate.get_capabilities_objects()[0]
      name = 'feature'
    else:
      self.nodeTemplate = nodeTemplate
      cap = nodeTemplate.get_capabilities()[name]
    self.type = nodeTemplate.type_definition.get_capabilities().get(name,
          CapabilityTypeDef(name, 'tosca.capabilities.Node', nodeTemplate.type))
    EntitySpec.__init__(self, cap)

  def isCompatibleTarget(self, targetStr):
    if self.name == targetStr:
      return True
    return self.type.is_derived_from(targetStr)

  def isCompatibleType(self, typeStr):
    return self.type.is_derived_from(typeStr)

  def getInterfaces(self):
    # capabilities don't have their own interfaces
    return self.nodeTemplate.interfaces



import os.path
import sys
import six
import codecs
from six.moves import urllib

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap
from ruamel.yaml.representer import RepresenterError

from .util import sensitive_str, UnfurlError, UnfurlValidationError
from .util import expandDoc, findSchemaErrors, makeMapWithBase
from toscaparser.common.exception import ExceptionCollector
from toscaparser.common.exception import URLException
from toscaparser.utils.gettextutils import _

import logging
logger = logging.getLogger('unfurl')
yaml = YAML()
# monkey patch for better error message
def represent_undefined(self, data):
    raise RepresenterError('cannot represent an object: %s of type %s' % (data, type(data)))
yaml.representer.represent_undefined = represent_undefined
yaml.representer.add_representer(None, represent_undefined)

def represent_sensitive(dumper, data):
  return dumper.represent_scalar(u'tag:yaml.org,2002:str', "[[REDACTED]]")
yaml.representer.add_representer(sensitive_str, represent_sensitive)

def load_yaml(path, isFile=True, importLoader=None):
  try:
    originalPath = path
    bare = False
    manifest = importLoader and getattr(importLoader.tpl, 'manifest', None)
    # check if this path is into a git repo
    if manifest:
      repo, filePath, revision, bare = manifest.findRepoFromGitUrl(path, isFile, importLoader, True)
      if repo: # it's a git repo
        if bare:
          return yaml.load(repo.show(filePath, revision))
        # find the project that the loading file is in
        # tracks which commit was used, returns a workingDir
        workingDir = repo.checkout(revision)
        path = os.path.join(workingDir, filePath)
        isFile = True
      else:
        # if it's a file, check if it's only in the repo
        if isFile:
          repo, filePath, revision, bare = manifest.findPathInRepos(path, importLoader, True)
          if repo:
            if bare:
              return yaml.load(repo.show(filePath, revision))
            else:
              path = os.path.join(repo.workingDir, filePath)

    # XXX urls and files outside of a repo should be saved and commited to the repo the importLoader is in
    f = None
    try:
        f = codecs.open(path, encoding='utf-8', errors='strict') if isFile \
            else urllib.request.urlopen(path)
    except urllib.error.URLError as e:
        if hasattr(e, 'reason'):
            msg = (_('Failed to reach server "%(path)s". Reason is: '
                     '%(reason)s.')
                   % {'path': path, 'reason': e.reason})
            ExceptionCollector.appendException(URLException(what=msg))
            return
        elif hasattr(e, 'code'):
            msg = (_('The server "%(path)s" couldn\'t fulfill the request. '
                     'Error code: "%(code)s".')
                   % {'path': path, 'code': e.code})
            ExceptionCollector.appendException(URLException(what=msg))
            return
    except Exception as e:
        raise
    return yaml.load(f.read())
  except:
    if bare:
      msg = 'Could not retrieve %s from repo (originally %s)' % (filePath, originalPath)
    elif path != originalPath:
      msg = 'Could not load "%s" (originally "%s")' % (path, originalPath)
    else:
      msg = 'Could not load "%s"' % path
    raise UnfurlError(msg, True)

import toscaparser.imports
toscaparser.imports.YAML_LOADER = load_yaml

class YamlConfig(object):
  def __init__(self, config=None, path=None, validate=True, schema=None, loadHook=None):
    try:
      self.schema = schema
      if path:
        self.path = os.path.abspath(path)
        if os.path.isfile(self.path):
          with open(self.path, 'r') as f:
            config = f.read()
        # otherwise use default config
      else:
        self.path = None

      if isinstance(config, six.string_types):
        if self.path:
          # set name on a StringIO so parsing error messages include the path
          config = six.StringIO(config)
          config.name = self.path
        self.config = yaml.load(config)
      elif isinstance(config, dict):
        self.config = CommentedMap(config.items())
      else:
        self.config = config
      if not isinstance(self.config, CommentedMap):
        raise UnfurlValidationError('invalid YAML document: %s' % self.config)

      self._cachedDocIncludes = {}
      #schema should include defaults but can't validate because it doesn't understand includes
      #but should work most of time
      self.config.loadTemplate = self.loadInclude
      self.loadHook = loadHook

      self.baseDirs = [self.getBaseDir()]
      self.includes, expandedConfig = expandDoc(self.config, cls=makeMapWithBase(self.baseDirs[0]))
      self.expanded = expandedConfig
      # print('expanded')
      # yaml.dump(config, sys.stdout)
      errors = schema and self.validate(expandedConfig)
      if errors and validate:
        (message, schemaErrors) = errors
        raise UnfurlValidationError("JSON Schema validation failed: " + message, errors)
      else:
        self.valid = not not errors
    except:
      if self.path:
        msg = "Unable to load yaml config at %s" % self.path
      else:
        msg = "Unable to parse yaml config"
      raise UnfurlError(msg, True)

  def loadYaml(self, path, baseDir=None):
    path = os.path.abspath(os.path.join(baseDir or self.getBaseDir(), path))
    with open(path, 'r') as f:
      config = yaml.load(f)
    return path, config

  def getBaseDir(self):
    if self.path:
      return os.path.dirname(self.path)
    else:
      return '.'

  def dump(self, out=sys.stdout):
    yaml.dump(self.config, out)

  def validate(self, config):
    return findSchemaErrors(config, self.schema)

  def loadInclude(self, templatePath):
    if templatePath == self.baseDirs[-1]:
      self.baseDirs.pop()
      return

    if isinstance(templatePath, dict):
      value = templatePath.get('merge')
      if 'file' not in templatePath:
        raise UnfurlError('file missing from document %%include: %s' % templatePath)
      key = templatePath['file']
    else:
      value = None
      key = templatePath

    if key in self._cachedDocIncludes:
      path, template = self._cachedDocIncludes[key]
      baseDir = os.path.dirname(path)
      self.baseDirs.append(baseDir)
      return value, template, baseDir

    try:
      baseDir = self.baseDirs[-1]
      if self.loadHook:
        path, template = self.loadHook(self, templatePath, baseDir)
      else:
        path, template = self.loadYaml(key, baseDir)
      newBaseDir = os.path.dirname(path)
    except:
      raise UnfurlError('unable to load document %%include: %s (base: %s)' % (templatePath, baseDir), True, True)
    self.baseDirs.append(newBaseDir)

    self._cachedDocIncludes[key] = [path, template]
    return value, template, newBaseDir

def loadFromRepo(import_name, import_uri_def, basePath, repositories, manifest):
  """
  Returns (url or fullpath, parsed yaml)
  """
  context = CommentedMap(import_uri_def.items())
  context['base'] = basePath
  context['repositories'] = repositories
  context.manifest = manifest
  uridef = {k: v for k, v in import_uri_def.items() if k in ['file', 'repository']}
  # _load_import_template will invoke load_yaml above
  loader = toscaparser.imports.ImportsLoader(None, basePath, tpl=context)
  return (loader._load_import_template(import_name, uridef))



# Copyright 2019, David Wilson
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import
import os
import signal
import threading

import mitogen.core
import ansible_mitogen.affinity
import ansible_mitogen.loaders
import ansible_mitogen.mixins
import ansible_mitogen.process

import ansible
import ansible.executor.process.worker


ANSIBLE_VERSION_MIN = '2.3'
ANSIBLE_VERSION_MAX = '2.7'
NEW_VERSION_MSG = (
    "Your Ansible version (%s) is too recent. The most recent version\n"
    "supported by Mitogen for Ansible is %s.x. Please check the Mitogen\n"
    "release notes to see if a new version is available, otherwise\n"
    "subscribe to the corresponding GitHub issue to be notified when\n"
    "support becomes available.\n"
    "\n"
    "    https://mitogen.rtfd.io/en/latest/changelog.html\n"
    "    https://github.com/dw/mitogen/issues/\n"
)
OLD_VERSION_MSG = (
    "Your version of Ansible (%s) is too old. The oldest version supported by "
    "Mitogen for Ansible is %s."
)


def _assert_supported_release():
    """
    Throw AnsibleError with a descriptive message in case of being loaded into
    an unsupported Ansible release.
    """
    v = ansible.__version__

    if v[:len(ANSIBLE_VERSION_MIN)] < ANSIBLE_VERSION_MIN:
        raise ansible.errors.AnsibleError(
            OLD_VERSION_MSG % (v, ANSIBLE_VERSION_MIN)
        )

    if v[:len(ANSIBLE_VERSION_MAX)] > ANSIBLE_VERSION_MAX:
        raise ansible.errors.AnsibleError(
            NEW_VERSION_MSG % (ansible.__version__, ANSIBLE_VERSION_MAX)
        )


def _patch_awx_callback():
    """
    issue #400: AWX loads a display callback that suffers from thread-safety
    issues. Detect the presence of older AWX versions and patch the bug.
    """
    # AWX uses sitecustomize.py to force-load this package. If it exists, we're
    # running under AWX.
    try:
        from awx_display_callback.events import EventContext
        from awx_display_callback.events import event_context
    except ImportError:
        return

    if hasattr(EventContext(), '_local'):
        # Patched version.
        return

    def patch_add_local(self, **kwargs):
        tls = vars(self._local)
        ctx = tls.setdefault('_ctx', {})
        ctx.update(kwargs)

    EventContext._local = threading.local()
    EventContext.add_local = patch_add_local

_patch_awx_callback()


def wrap_action_loader__get(name, *args, **kwargs):
    """
    While the mitogen strategy is active, trap action_loader.get() calls,
    augmenting any fetched class with ActionModuleMixin, which replaces various
    helper methods inherited from ActionBase with implementations that avoid
    the use of shell fragments wherever possible.

    This is used instead of static subclassing as it generalizes to third party
    action modules outside the Ansible tree.
    """
    klass = action_loader__get(name, class_only=True)
    if klass:
        bases = (ansible_mitogen.mixins.ActionModuleMixin, klass)
        adorned_klass = type(str(name), bases, {})
        if kwargs.get('class_only'):
            return adorned_klass
        return adorned_klass(*args, **kwargs)


def wrap_connection_loader__get(name, *args, **kwargs):
    """
    While the strategy is active, rewrite connection_loader.get() calls for
    some transports into requests for a compatible Mitogen transport.
    """
    if name in ('docker', 'kubectl', 'jail', 'local', 'lxc',
                'lxd', 'machinectl', 'setns', 'ssh'):
        name = 'mitogen_' + name
    return connection_loader__get(name, *args, **kwargs)


def wrap_worker__run(*args, **kwargs):
    """
    While the strategy is active, rewrite connection_loader.get() calls for
    some transports into requests for a compatible Mitogen transport.
    """
    # Ignore parent's attempts to murder us when we still need to write
    # profiling output.
    if mitogen.core._profile_hook.__name__ != '_profile_hook':
        signal.signal(signal.SIGTERM, signal.SIG_IGN)

    ansible_mitogen.logging.set_process_name('task')
    ansible_mitogen.affinity.policy.assign_worker()
    return mitogen.core._profile_hook('WorkerProcess',
        lambda: worker__run(*args, **kwargs)
    )


class StrategyMixin(object):
    """
    This mix-in enhances any built-in strategy by arranging for various Mitogen
    services to be initialized in the Ansible top-level process, and for worker
    processes to grow support for using those top-level services to communicate
    with and execute modules on remote hosts.

    Mitogen:

        A private Broker IO multiplexer thread is created to dispatch IO
        between the local Router and any connected streams, including streams
        connected to Ansible WorkerProcesses, and SSH commands implementing
        connections to remote machines.

        A Router is created that implements message dispatch to any locally
        registered handlers, and message routing for remote streams. Router is
        the junction point through which WorkerProceses and remote SSH contexts
        can communicate.

        Router additionally adds message handlers for a variety of base
        services, review the Standard Handles section of the How It Works guide
        in the documentation.

        A ContextService is installed as a message handler in the master
        process and run on a private thread. It is responsible for accepting
        requests to establish new SSH connections from worker processes, and
        ensuring precisely one connection exists and is reused for subsequent
        playbook steps. The service presently runs in a single thread, so to
        begin with, new SSH connections are serialized.

        Finally a mitogen.unix listener is created through which WorkerProcess
        can establish a connection back into the master process, in order to
        avail of ContextService. A UNIX listener socket is necessary as there
        is no more sane mechanism to arrange for IPC between the Router in the
        master process, and the corresponding Router in the worker process.

    Ansible:

        PluginLoader monkey patches are installed to catch attempts to create
        connection and action plug-ins.

        For connection plug-ins, if the desired method is "local" or "ssh", it
        is redirected to the "mitogen" connection plug-in. That plug-in
        implements communication via a UNIX socket connection to the top-level
        Ansible process, and uses ContextService running in the top-level
        process to actually establish and manage the connection.

        For action plug-ins, the original class is looked up as usual, but a
        new subclass is created dynamically in order to mix-in
        ansible_mitogen.target.ActionModuleMixin, which overrides many of the
        methods usually inherited from ActionBase in order to replace them with
        pure-Python equivalents that avoid the use of shell.

        In particular, _execute_module() is overridden with an implementation
        that uses ansible_mitogen.target.run_module() executed in the target
        Context. run_module() implements module execution by importing the
        module as if it were a normal Python module, and capturing its output
        in the remote process. Since the Mitogen module loader is active in the
        remote process, all the heavy lifting of transferring the action module
        and its dependencies are automatically handled by Mitogen.
    """
    def _install_wrappers(self):
        """
        Install our PluginLoader monkey patches and update global variables
        with references to the real functions.
        """
        global action_loader__get
        action_loader__get = ansible_mitogen.loaders.action_loader.get
        ansible_mitogen.loaders.action_loader.get = wrap_action_loader__get

        global connection_loader__get
        connection_loader__get = ansible_mitogen.loaders.connection_loader.get
        ansible_mitogen.loaders.connection_loader.get = wrap_connection_loader__get

        global worker__run
        worker__run = ansible.executor.process.worker.WorkerProcess.run
        ansible.executor.process.worker.WorkerProcess.run = wrap_worker__run

    def _remove_wrappers(self):
        """
        Uninstall the PluginLoader monkey patches.
        """
        ansible_mitogen.loaders.action_loader.get = action_loader__get
        ansible_mitogen.loaders.connection_loader.get = connection_loader__get
        ansible.executor.process.worker.WorkerProcess.run = worker__run

    def _add_plugin_paths(self):
        """
        Add the Mitogen plug-in directories to the ModuleLoader path, avoiding
        the need for manual configuration.
        """
        base_dir = os.path.join(os.path.dirname(__file__), 'plugins')
        ansible_mitogen.loaders.connection_loader.add_directory(
            os.path.join(base_dir, 'connection')
        )
        ansible_mitogen.loaders.action_loader.add_directory(
            os.path.join(base_dir, 'action')
        )

    def _queue_task(self, host, task, task_vars, play_context):
        """
        Many PluginLoader caches are defective as they are only populated in
        the ephemeral WorkerProcess. Touch each plug-in path before forking to
        ensure all workers receive a hot cache.
        """
        ansible_mitogen.loaders.module_loader.find_plugin(
            name=task.action,
            mod_type='',
        )
        ansible_mitogen.loaders.connection_loader.get(
            name=play_context.connection,
            class_only=True,
        )
        ansible_mitogen.loaders.action_loader.get(
            name=task.action,
            class_only=True,
        )

        return super(StrategyMixin, self)._queue_task(
            host=host,
            task=task,
            task_vars=task_vars,
            play_context=play_context,
        )

    def run(self, iterator, play_context, result=0):
        """
        Arrange for a mitogen.master.Router to be available for the duration of
        the strategy's real run() method.
        """
        _assert_supported_release()

        ansible_mitogen.process.MuxProcess.start()
        run = super(StrategyMixin, self).run
        self._add_plugin_paths()
        self._install_wrappers()
        try:
            return mitogen.core._profile_hook('Strategy',
                lambda: run(iterator, play_context)
            )
        finally:
            self._remove_wrappers()



import unittest
import os
import traceback
from collections import MutableSequence
from click.testing import CliRunner
from unfurl.__main__ import cli
from unfurl import __version__
from unfurl.configurator import Configurator

manifest = """
apiVersion: unfurl/v1alpha1
kind: Manifest
imports:
  local:
    properties:
      prop2:
       type: number
spec:
  node_templates:
    test:
      type: tosca.nodes.Root
      properties:
        local1:
          eval:
            local: prop1
        local2:
          eval:
            external: local
          foreach: prop2
        testApikey:
          eval:
           secret: testApikey
        # uses jinja2 native types to evaluate to a list
        aListOfItems:
            eval:
              template: >
                [{% for key, value in aLocalDict.items() %}
                  {
                    'key': '{{ key }}',
                    'value': '{{ value | b64encode }}'
                  },
                {% endfor %}]
            vars:
              aLocalDict:
               eval:
                # XXX test with secret instead (need to propagate secretness)
                secret: aDict
            # trace: 1
      interfaces:
        Standard:
          create: test_cli.CliTestConfigurator
"""

localConfig = """
defaults: #used if manifest isnt found in `manifests` list below
 secret:
  attributes:
    aDict:
      key1: a string
      key2: 2
    default: # if key isn't found, apply this:
      q: # quote
        eval:
          lookup:
            env: "GEO_{{ key | upper }}"

instances:
  - file: git/default-manifest.yaml
    local:
      attributes:
        prop1: 'found'
        prop2: 1
        aDict:
          key1: a string
          key2: 2
"""

class CliTestConfigurator(Configurator):
  def run(self, task):
    attrs = task.target.attributes
    assert isinstance(attrs['aListOfItems'], MutableSequence), type(attrs['aListOfItems'])
    # sort for python2
    assert sorted(attrs['aListOfItems'], key=lambda k: k['key']) == [{
        'key': 'key1',
        'value': 'YSBzdHJpbmc='
      },
      {
        'key': 'key2',
        'value': 'Mg=='
      }], attrs['aListOfItems']
    assert attrs['local1'] == 'found', attrs['local1']
    assert attrs['local2'] == 1, attrs['local2']
    assert attrs['testApikey'] == 'secret', 'failed to get secret environment variable, maybe DelegateAttributes is broken?'
    assert attrs._attributes['aListOfItems'].external.type == 'sensitive'
    # XXX this should be marked as secret so it should serializes as "[REDACTED]"
    # attrs['copyofAListOfItems'] = attrs['aListOfItems']
    yield task.createResult(True, False, "ok")

class CliTest(unittest.TestCase):

  def test_help(self):
    runner = CliRunner()
    result = runner.invoke(cli, [])
    assert result.output.startswith("Usage: cli [OPTIONS] COMMAND [ARGS]"), result.output
    self.assertEqual(result.exit_code, 0)

  def test_version(self):
    runner = CliRunner()
    result = runner.invoke(cli, ['version'])
    self.assertEqual(result.exit_code, 0, result)
    self.assertEqual(result.output.strip(), "unfurl version %s" % __version__)

  def test_badargs(self):
    runner = CliRunner()
    result = runner.invoke(cli, ['--badarg'])
    self.assertEqual(result.exit_code, 2, result)

  def test_run(self):
    runner = CliRunner()
    with runner.isolated_filesystem():
      with open('manifest.yaml', 'w') as f:
        f.write('invalid manifest')

      result = runner.invoke(cli, ['run'])
      self.assertEqual(result.exit_code, 1, result)
      # XXX log handler is writing to the CliRunner's output stream
      self.assertEqual(result.output.strip(), 'Unable to create job')

  def test_localConfig(self):
    # test loading the default manifest declared in the local config
    # test locals and secrets:
    #    declared attributes and default lookup
    #    inherited from default (inheritFrom)
    #    verify secret contents isn't saved in config
    os.environ['GEO_TESTAPIKEY'] = 'secret'
    runner = CliRunner()
    with runner.isolated_filesystem() as tempDir:
      with open('unfurl.yaml', 'w') as local:
        local.write(localConfig)
      repoDir = 'git'
      os.mkdir(repoDir)
      os.chdir(repoDir)
      with open('default-manifest.yaml', 'w') as f:
        f.write(manifest)
      result = runner.invoke(cli, ['-vvv', 'deploy', '--jobexitcode', 'degraded'])
      # uncomment this to see output:
      print("result.output", result.exit_code, result.output)
      assert not result.exception, '\n'.join(traceback.format_exception(*result.exc_info))
      self.assertEqual(result.exit_code, 0, result.output)



"""
Public Api:

mapValue - returns a copy of the given value resolving any embedded queries or template strings
Ref.resolve given an expression, returns a ResultList
Ref.resolveOne given an expression, return value, none or a (regular) list
Ref.isRef return true if the given diction looks like a Ref

Internal:
evalRef() given expression (string or dictionary) return list of Results
Expr.resolve() given expression string, return list of Results
Results._mapValue same as mapValue but with lazily evaluation
"""
import six
import re
import operator
import collections
from collections import Mapping, MutableSequence
from ruamel.yaml.comments import CommentedMap
from .util import validateSchema, UnfurlError
from .result import ResultsList, Result, Results, ExternalValue, ResourceRef

def mapValue(value, resourceOrCxt):
  if not isinstance(resourceOrCxt, RefContext):
    resourceOrCxt = RefContext(resourceOrCxt)
  return _mapValue(value, resourceOrCxt)

def _mapValue(value, ctx, wantList=False):
  from .support import isTemplate, applyTemplate

  if Ref.isRef(value):
    # wantList=False := resolveOne
    return Ref(value).resolve(ctx, wantList=wantList)

  if isinstance(value, Mapping):
    try:
      oldBaseDir = ctx.baseDir
      ctx.baseDir = getattr(value, 'baseDir', oldBaseDir)
      if ctx.baseDir and ctx.baseDir != oldBaseDir:
        ctx.trace("found baseDir", ctx.baseDir)
      return dict((key, _mapValue(v, ctx, wantList)) for key, v in value.items())
    finally:
      ctx.baseDir = oldBaseDir
  elif isinstance(value, (MutableSequence, tuple)):
    return [_mapValue(item, ctx, wantList) for item in value]
  elif isTemplate(value, ctx):
    return applyTemplate(value, ctx)
  return value

class _Tracker(object):
  def __init__(self):
    self.count = 0
    self.referenced = []

  def start(self):
    self.count += 1
    return len(self.referenced)

  def stop(self):
    self.count -= 1
    return self.count

  def addReference(self, ref, result):
    if self.count > 0:
      self.referenced.append([ref, result])

  def getReferencedResults(self, index=0):
    referenced = self.referenced[index:]
    if not referenced:
      return
    for (ref, results) in referenced:
      if not isinstance(ref, Ref):
        assert isinstance(results, Result)
        yield results
      else:
        for obj in results._attributes:
          assert isinstance(obj, Result)
          yield obj

class RefContext(object):
  def __init__(self, currentResource, vars=None, wantList=False, resolveExternal=False, trace=0):
    self.vars = vars or {}
    # the original context:
    self.currentResource = currentResource
    # the last resource encountered while evaluating:
    self._lastResource = currentResource
    # current segment is the final segment:
    self._rest = None
    self.wantList = wantList
    self.resolveExternal = resolveExternal
    self._trace = trace
    self.baseDir = currentResource.root.baseDir
    self.templar = currentResource.templar
    self.referenced = _Tracker()

  def copy(self, resource=None, vars=None, wantList=None, trace=0):
    copy = RefContext(resource or self.currentResource, self.vars, self.wantList,
      self.resolveExternal, max(self._trace, trace))
    if not isinstance(copy.currentResource, ResourceRef) and isinstance(self._lastResource, ResourceRef):
      copy._lastResource = self._lastResource
    if vars:
      copy.vars.update(vars)
    if wantList is not None:
      copy.wantList = wantList
    copy.baseDir = self.baseDir
    copy.templar = self.templar
    copy.referenced = self.referenced
    return copy

  def trace(self, *msg):
    if self._trace:
      print("%s (ctx: %s)" % (' '.join(str(a) for a in msg), self._lastResource))

  def addReference(self, ref, result):
    self.referenced.addReference(ref, result)

  def resolveVar(self, key):
    value = self.vars[key[1:]]
    if isinstance(value, Result):
      return value.resolved
    else:
      return value

class Expr(object):
  def __init__(self, exp, vars = None):
    self.vars = {
     'true': True, 'false': False, 'null': None
    }

    if vars:
      self.vars.update(vars)

    self.source = exp
    paths = list(parseExp(exp))
    if 'break' not in self.vars:
      # hack to check that we aren't a foreach expression
      if (not paths[0].key or paths[0].key[0] not in '.$') and (paths[0].key or paths[0].filters):
        # unspecified relative path: prepend a segment to select ancestors
        # and have the first segment only choose the first match
        # note: the first match modifier is set on the first segment and not .ancestors
        # because .ancestors is evaluated against the initial context resource and so
        # its first match is always going to be the result of the whole query (since the initial context is just one item)
        paths[:1] = [Segment('.ancestors', [], '', []), paths[0]._replace(modifier='?')]
    self.paths = paths

  def __repr__(self):
    # XXX vars
    return "Expr('%s')" % self.source

  def resolve(self, context):
    # returns a list of Results
    currentResource = context.currentResource
    if not self.paths[0].key and not self.paths[0].filters: # starts with "::"
      currentResource = currentResource.all
      paths = self.paths[1:]
    elif self.paths[0].key and self.paths[0].key[0] == '$':
      #if starts with a var, use that as the start
      currentResource = context.resolveVar(self.paths[0].key)
      if len(self.paths) == 1:
        # bare reference to a var, just return it's value
        return [Result(currentResource)]
      paths = [self.paths[0]._replace(key='')] + self.paths[1:]
    else:
      paths = self.paths
    return evalExp([currentResource], paths, context)

class Ref(object):
  """
  A Ref objects describes a path to metadata associated with a resource.

  The syntax for a Ref path expression is:

  expr:  segment? ('::' segment)*

  segment: key? ('[' filter ']')* '?'?

  key: name | integer | var | '*'

  filter: '!'? expr? (('!=' | '=') test)?

  test: var | (^[$[]:?])*

  var: '$' name

  Semantics

  Each segment specifies a key in a resource or JSON/YAML object.
  "::" is used as the segment deliminated to allow for keys that contain "." and "/"

  Path expressions evaluations always start with a list of one or more Resources.
  and each segment selects the value associated with that key.
  If segment has one or more filters
  each filter is applied to that value -- each is treated as a predicate
  that decides whether value is included or not in the results.
  If the filter doesn't include a test the filter tests the existence or non-existence of the expression,
  depending on whether the expression is prefixed with a "!".
  If the filter includes a test the left side of the test needs to match the right side.
  If the right side is not a variable, that string will be coerced to left side's type before comparing it.
  If the left-side expression is omitted, the value of the segment's key is used and if that is missing, the current value is used.

  If the current value is a list and the key looks like an integer
  it will be treated like a zero-based index into the list.
  Otherwise the segment is evaluated again all values in the list and resulting value is a list.
  If the current value is a dictionary and the key is "*", all values will be selected.

  If a segment ends in "?", it will only include the first match.
  In other words, "a?::b::c" is a shorthand for "a[b::c]::0::b::c".
  This is useful to guarantee the result of evaluating expression is always a single result.

  The first segment:
  If the first segment is a variable reference the current value is set to that variable's value.
  If the key in the first segment is empty (e.g. the expression starts with '::') the current value will be set to the evaluation of '.all'.
  If the key in the first segment starts with '.' it is evaluated against the initial "current resource".
  Otherwise, the current value is set to the evaluation of ".ancestors?". In other words,
  the expression will be the result of evaluating it against the first ancestor of the current resource that it matches.

  If key or test needs to be a non-string type or contains a unallowed character use a var reference instead.

  When multiple steps resolve to lists the resultant lists are flattened.
  However if the final set of matches contain values that are lists those values are not flattened.

  For example, given:

  .. code-block:: javascript

    {x: [ {
            a: [{c:1}, {c:2}]
          },
          {
            a: [{c:3}, {c:4}]
          }
        ]
    }

  `x:a:c` resolves to:
    `[1,2,3,4]`
  not
    `[[1,2], [3,4]])`

  (Justification: It is inconvenient and fragile to tie data structures to the particular form of a query.
  If you want preserve structure (e.g. to know which values are part
  of which parent value or resource) use a less deep path and iterate over results.)

  Resources have a special set of keys:

  .            self
  ..           parent
  .parents     list of parents
  .ancestors   self and parents
  .root        root ancestor
  .children    child resources
  .descendents (including self)
  .all       dictionary of child resources with their names as keys
  .configurations
  """

  def __init__(self, exp, vars = None):
    self.vars = {
     'true': True, 'false': False, 'null': None
    }

    self.foreach = None
    self.trace = 0
    if isinstance(exp, Mapping):
      keys = list(exp)
      if keys and keys[0] not in _FuncsTop:
        self.vars.update(exp.get('vars', {}))
        self.foreach = exp.get('foreach')
        self.trace = exp.get('trace', 0)
        exp = exp.get('eval', exp.get('ref', exp))

    if vars:
      self.vars.update(vars)
    self.source = exp

  def resolve(self, ctx, wantList=True):
    """
    If wantList=True (default) returns a ResultList of matches
    Note that values in the list can be a list or None
    If wantList=False return `resolveOne` semantics
    If wantList='result' return a Result
    """
    ctx = ctx.copy(vars=self.vars, wantList=wantList, trace=self.trace)
    ctx.trace('Ref.resolve(wantList=%s) start' % wantList, self.source)
    results = evalRef(self.source, ctx, True)
    ctx.trace('Ref.resolve(wantList=%s) evalRef' % wantList, self.source, results)
    if results and self.foreach:
      results = forEach(self.foreach, results, ctx)
    assert not isinstance(results, ResultsList), results
    results = ResultsList(results, ctx)
    ctx.addReference(self, results)
    ctx.trace('Ref.resolve(wantList=%s) results' % wantList, self.source, results)
    if wantList and not wantList == 'result':
      return results
    else:
      if not results:
        return None
      elif len(results) == 1:
        if wantList == 'result':
          return results._attributes[0]
        else:
          return results[0]
      else:
        if wantList == 'result':
          return Result(results)
        else:
          return list(results)

  def resolveOne(self, ctx):
    """
    If no match return None
    If more than one match return a list of matches
    Otherwise return the match

    Note: If you want to distinguish between None values and no match
    or between single match that is a list and a list of matches
    use resolve() which always returns a (possible empty) of matches
    """
    return self.resolve(ctx, False)

  @staticmethod
  def isRef(value):
    if isinstance(value, Mapping):
      if 'ref' in value or 'eval' in value:
        return len([x for x in ['vars', 'trace', 'foreach'] if x in value]) + 1 == len(value)
      if len(value) == 1 and list(value)[0] in _FuncsTop:
        return True
      return False
    return isinstance(value, Ref)

def evalAsBoolean(arg, ctx):
  result = evalRef(arg, ctx)
  return not not result[0].resolved if result else False

def ifFunc(arg, ctx):
  kw = ctx.kw
  result = evalAsBoolean(arg, ctx)
  if result:
    return evalForFunc(kw.get('then'), ctx)
  else:
    return evalForFunc(kw.get('else'), ctx)

def orFunc(arg, ctx):
  args = evalForFunc(arg, ctx)
  assert isinstance(args, MutableSequence), args
  for arg in args:
    val = evalForFunc(arg, ctx)
    if val:
      return val

def notFunc(arg, ctx):
  result = evalAsBoolean(arg, ctx)
  return not result

def andFunc(arg, ctx):
  args = evalForFunc(arg, ctx)
  assert isinstance(args, MutableSequence)
  for arg in args:
    val = evalForFunc(arg, ctx)
    if not val:
      return val
  return val

def quoteFunc(arg, ctx):
  return arg

def eqFunc(arg, ctx):
  args = mapValue(arg, ctx)
  assert isinstance(args, MutableSequence) and len(args) == 2
  return args[0] == args[1]

def validateSchemaFunc(arg, ctx):
  args = mapValue(arg, ctx)
  assert isinstance(args, MutableSequence) and len(args) == 2
  return validateSchema(args[0], args[1])

def _forEach(foreach, results, ctx):
  if isinstance(foreach, six.string_types):
    keyExp = None
    valExp = foreach
  else:
    keyExp = foreach.get('key')
    valExp = foreach.get('value')
    if not valExp and not keyExp:
      # it's a dict that needs to be evaluated
      valExp = foreach

  ictx = ctx.copy(wantList=False)
  # ictx._trace = 1
  Break = object()
  Continue = object()
  def makeItems():
    for i, (k, v) in enumerate(results):
      ictx.currentResource = v
      ictx.vars['collection'] = results
      ictx.vars['index'] = i
      ictx.vars['key'] = k
      ictx.vars['item'] = v
      ictx.vars['break'] = Break
      ictx.vars['continue'] = Continue
      if keyExp:
        key = evalForFunc(keyExp, ictx)
        if key is Break:
          break
        elif key is Continue:
          continue
      valResults = evalRef(valExp, ictx)
      if not valResults:
        continue
      if len(valResults) == 1:
        val = valResults[0].resolved
        if val is Break:
          break
        elif val is Continue:
          continue
        valResults = valResults[0]

      if keyExp:
        yield (key, valResults)
      else:
        yield valResults

  if keyExp:
    # use CommentedMap to preserve order
    return [CommentedMap(makeItems())]
  else:
    return list(makeItems())

def forEach(foreach, results, ctx):
  # results will be list of Results
  return _forEach(foreach, enumerate(r.external or r.resolved for r in results), ctx)

def forEachFunc(foreach, ctx):
  results = ctx.currentResource
  if results:
    if isinstance(results, Mapping):
      return _forEach(foreach, results.items(), ctx)
    elif isinstance(results, MutableSequence):
      return _forEach(foreach, enumerate(results), ctx)
    else:
      return _forEach(foreach, [(0, results)], ctx)
  else:
    return results

_Funcs = {
  'if': ifFunc,
  'and': andFunc,
  'or': orFunc,
  'not': notFunc,
  'q': quoteFunc,
  'eq': eqFunc,
  'validate': validateSchemaFunc,
  'foreach': forEachFunc,
}
_FuncsTop = ['q']
def getEvalFunc(name):
  return _Funcs.get(name)

def setEvalFunc(name, val, topLevel = False):
  _Funcs[name] = val
  if topLevel:
    _FuncsTop.append(name)

# returns list of results
def evalRef(val, ctx, top=False):
  "val is assumed to be an expression, evaluate and return a list of Results"
  from .support import isTemplate, applyTemplate

  # functions and ResultsMap assume resolveOne semantics
  if top:
    varctx = ctx.copy()
    varctx.vars = {} # var context can't have itself as its vars
    vars = _mapValue(ctx.vars, varctx, wantList='result')
    vars['start'] = ctx.currentResource
    ctx = ctx.copy(ctx.currentResource, vars, wantList = False)

  if isinstance(val, Mapping):
    for key in val:
      func = _Funcs.get(key)
      if func:
        args = val[key]
        ctx.kw = val
        ctx.currentFunc = key
        if key != 'foreach' and 'foreach' in val:
          raise UnfurlError("unexpected 'foreach' found, did you intend it for the parent?")
        val = func(args, ctx)
        if key == 'q':
          return [Result(val)]
        break
  elif isinstance(val, six.string_types):
    if isTemplate(val, ctx):
      return [Result(applyTemplate(val, ctx))]
    else:
      expr = Expr(val, ctx.vars)
      results = expr.resolve(ctx) # returns a list of Results
      ctx.trace('expr.resolve', results)
      return results

  mappedVal = Results._mapValue(val, ctx)
  if isinstance(mappedVal, Result):
    return [mappedVal]
  else:
    return [Result(mappedVal)]

def evalForFunc(val, ctx):
  "like `evalRef` except it returns the resolved value"
  results = evalRef(val, ctx)
  if not results:
    return None
  if len(results) == 1:
    return results[0].resolved
  else:
    return [r.resolved for r in results]

#return a segment
Segment = collections.namedtuple('Segment', ['key', 'test', 'modifier', 'filters'])
defaultSegment = Segment('', [], '', [])

def evalTest(value, test, context):
  comparor = test[0]
  key = test[1]
  try:
    if context and isinstance(key, six.string_types) and key.startswith('$'):
      compare = context.resolveVar(key)
    else:
      # try to coerce string to value type
      compare = type(value)(key)
    context.trace('compare', value, compare, comparor(value, compare))
    if comparor(value, compare):
      return True
  except:
    context.trace('compare exception, ne:',  comparor is operator.ne)
    if comparor is operator.ne:
      return True
  return False

# given a Result and a key, return None or new Result
def lookup(result, key, context):
  try:
    # if key == '.':
    #   key = context.currentKey
    if context and isinstance(key, six.string_types) and key.startswith('$'):
      key = context.resolveVar(key)

    if isinstance(result.resolved, ResourceRef):
      context._lastResource = result.resolved

    ctx = context.copy(context._lastResource)
    result = result.project(key, ctx)
    value = result.resolved
    context.trace('lookup %s, got %s' % (key, value))

    if not context._rest:
      assert not Ref.isRef(value)
      result.resolved = Results._mapValue(value, ctx)
      assert not isinstance(result.resolved, (ExternalValue, Result))

    return result
  except (KeyError, IndexError, TypeError, ValueError):
    if context._trace:
      context.trace('lookup return None due to exception:')
      import traceback
      traceback.print_exc()
    return None

# given a Result, yields the result
def evalItem(result, seg, context):
  """
    apply current item to current segment, return [] or [value]
  """
  if seg.key != '':
    result = lookup(result, seg.key, context)
    if not result:
      return

  value = result.resolved
  for filter in seg.filters:
    if _treatAsSingular(result, filter[0]):
      resultList = [value]
    else:
      resultList = value
    results = evalExp(resultList, filter, context)
    negate = filter[0].modifier == '!'
    if negate and results:
      return
    elif not negate and not results:
      return

  if seg.test and not evalTest(value, seg.test, context):
    return
  assert isinstance(result, Result), result
  yield result

def _treatAsSingular(result, seg):
  if seg.key == '*':
    return False
  # treat external values as single item even if they resolve to a list
  # treat lists as a single item if indexing into it
  return result.external or not isinstance(result.resolved, MutableSequence) or isinstance(seg.key, six.integer_types)

def recursiveEval(v, exp, context):
  """
  given a iterator of (previous) results,
  yields Results
  """
  context.trace('recursive evaluating', exp)
  matchFirst = exp[0].modifier == '?'
  useValue = exp[0].key == '*'
  for result in v:
    assert isinstance(result, Result), result
    item = result.resolved

    if _treatAsSingular(result, exp[0]):
      rest = exp[1:]
      context._rest = rest
      context.trace('evaluating item %s with key %s' % (item, exp[0].key))
      iv = evalItem(result, exp[0], context) # returns a generator that yields up to one result
    else:
      iv = result._values()
      if useValue:
        if not isinstance(item, Mapping):
          context.trace('* is skipping', item)
          continue
        rest = exp[1:] # advance past "*" segment
      else:
        # flattens
        rest = exp
        context.trace('flattening', item)

    # iv will be a generator or list
    if rest:
      results = recursiveEval(iv, rest, context)
      found = False
      for r in results:
        found = True
        context.trace('recursive result', r)
        assert isinstance(r, Result)
        yield r
      context.trace('found recursive %s matchFirst: %s' % (found, matchFirst))
      if found and matchFirst:
        return
    else:
      for r in iv:
        assert isinstance(r, Result)
        yield r
        if matchFirst:
          return

def evalExp(start, paths, context):
  "Returns a list of Results"
  context.trace('evalexp', start, paths)
  assert isinstance(start, MutableSequence), start
  return list(recursiveEval((Result(i) for i in start), paths, context))

def _makeKey(key):
  try:
    return int(key)
  except ValueError:
    return key

def parsePathKey(segment):
  #key, negation, test, matchFirst
  if not segment:
    return defaultSegment

  modifier = ''
  if segment[0] == '!':
    segment = segment[1:]
    modifier = '!'
  elif segment[-1] == '?':
    segment = segment[:-1]
    modifier = '?'

  parts = re.split(r'(=|!=)', segment, 1)
  if len(parts) == 3:
    key = parts[0]
    op = operator.eq if parts[1] == '=' else operator.ne
    return Segment(_makeKey(key), [op, parts[2]], modifier, [])
  else:
    return Segment(_makeKey(segment), [], modifier, [])

def parsePath(path, start):
  paths = path.split('::')
  segments = [parsePathKey(k.strip()) for k in paths]
  if start:
    if paths and paths[0]:
      # if the path didn't start with ':' merge with the last segment
      # e.g. foo[]? d=test[d]?
      segments[0] = start._replace(test=segments[0].test or start.test,
                    modifier=segments[0].modifier or start.modifier)
    else:
      return [start] + segments
  return segments

def parseExp(exp):
  #return list of steps
  rest = exp
  last = None

  while rest:
    steps, rest = parseStep(rest, last)
    last = None
    if steps:
      #we might need merge the next step into the last
      last = steps.pop()
      for step in steps:
        yield step

  if last:
    yield last

def parseStep(exp, start=None):
  split = re.split(r'(\[|\])', exp, 1)
  if len(split) == 1: #not found
    return parsePath(split[0], start), ''
  else:
    path, sep, rest = split

  paths = parsePath(path, start)

  filterExps = []
  while sep == '[':
    filterExp, rest = parseStep(rest)
    filterExps.append(filterExp)
    #rest will be anything after ]
    sep = rest and rest[0]

  #add filterExps to last Segment
  paths[-1] = paths[-1]._replace(filters = filterExps)
  return paths, rest



"""
A Job is generated by comparing a list of specs with the last known state of the system.
Job runs tasks, each of which has a configuration spec that is executed on the running system
Each task tracks and records its modifications to the system's state
"""

import collections
import datetime
import types
from .support import Status, Priority, AttributeManager
from .result import serializeValue, ChangeRecord
from .util import UnfurlError, UnfurlTaskError, mergeDicts, ansibleDisplay, toEnum
from .runtime import OperationalInstance
from .configurator import TaskView, ConfiguratorResult
from .plan import Plan

import logging
logger = logging.getLogger('unfurl')

class ConfigChange(OperationalInstance, ChangeRecord):
  """
  Represents a configuration change made to the system.
  It has a operating status and a list of dependencies that contribute to its status.
  There are two kinds of dependencies:
    1. Live resource attributes that the configuration's inputs depend on.
    2. Other configurations and resources it relies on to function properly.
  """
  def __init__(self, status=None, **kw):
    OperationalInstance.__init__(self, status, **kw)
    ChangeRecord.__init__(self)

class TaskRequest(object):
  def __init__(self, configSpec, resource, persist, required):
    self.configSpec = configSpec
    self.target = resource
    self.persist = persist
    self.required = required

class JobRequest(object):
  def __init__(self, resources, errors):
    self.resources = resources
    self.errors = errors

class ConfigTask(ConfigChange, TaskView, AttributeManager):
  """
    receives a configSpec and a target node instance
    instantiates and runs Configurator
    updates Configurator's target's status and lastConfigChange
  """
  def __init__(self, job, configSpec, target, parentId=None, reason = None):
    ConfigChange.__init__(self, lastConfigChange=target.lastConfigChange)
    TaskView.__init__(self, job.runner.manifest, configSpec, target, reason)
    AttributeManager.__init__(self)
    self.parentId = parentId or job.changeId
    self.changeId = self.parentId
    self.startTime = job.startTime or datetime.datetime.now()
    self.errors = []
    self.dryRun = job.dryRun
    self.generator = None
    self.job = job
    self.changeList = []
    self.result = None

    # set the attribute manager on the root resource
    # XXX refcontext in attributeManager should define $TARGET $HOST etc.
    #self.configuratorResource.root.attributeManager = self
    self.target.root.attributeManager = self

    # XXX set self.configSpec.target.readyState = 'pending'
    self.configurator = self.configSpec.create()

  def priority():
    doc = "The priority property."
    def fget(self):
      if self._priority is None:
        return self.configSpec.shouldRun()
      else:
        return self._priority
    def fset(self, value):
      self._priority = value
    def fdel(self):
      del self._priority
    return locals()
  priority = property(**priority())

  def startRun(self):
    if self.dryRun:
      self.generator = self.configurator.dryRun(self)
    else:
      self.generator = self.configurator.run(self)
    assert isinstance(self.generator, types.GeneratorType)

  def send(self, change):
    result = None
    try:
      result = self.generator.send(change)
    finally:
      # serialize configuration changes
      self.commitChanges()
    return result

  def start(self):
    self.startRun()

  def _setConfigStatus(self, config, result):
    statusChanged = config.localStatus != result.readyState
    if result.configChanged is None:
      # not set so try to deduce
      if self.reason in ['config changed', 'all']:
        configChanged = statusChanged or self.changeList or self.dependenciesChanged
      else: # be conservative, assume the worse
        configChanged = True
    else:
      # setting result.configChanged will override change detection
      configChanged = result.configChanged

    if configChanged:
      config._lastConfigChange = self.changeId
    if statusChanged:
      config.localStatus = result.readyState
    logger.debug("task %s statusChanged: %s configChanged %s", self, statusChanged and config.localStatus, configChanged and config._lastConfigChange)

  def processResult(self, result):
    """
    Update the target instance with the result.

    `result.applied` indicates this configuration is active
    (essentially, the owner of the instance's configuration)
    `result.modified` indicates if a "physical" change to this system was made.
     (All combinations of these two are permissible -- modified and not applied
     means changes were made to the system that couldn't be undone.
    """
    instance = self.target
    if result.modified:
      instance._lastStateChange = self.changeId

    if result.applied:
      assert result.readyState and result.readyState != Status.notapplied, result.readyState
      self._setConfigStatus(instance, result)
    else:
      if instance._lastConfigChange is None:
        # if this has never been set before, record this to indicate we tried
        instance._lastConfigChange = self.changeId
      # configurator wasn't able to apply so leave the instance state as is
      # except in the case where explicitly set another status
      # e.g. if it left the instance in an error state
      if result.readyState and result.readyState != Status.notapplied:
        instance.localStatus = result.readyState

  def finished(self, result):
    assert result
    if self.generator:
      self.generator.close()
      self.generator = None

    # don't set the changeId until we're finish so that we have a higher changeid
    # than nested tasks and jobs that ran (avoids spurious config changed tasks)
    self.changeId = self.job.runner.incrementChangeId()
    #XXX2 if attributes changed validate using attributesSchema
    #XXX2 Check that configuration provided the metadata that it declared (check postCondition)
    self.processResult(result)
    resource = self.target

    if (result.applied or result.modified) and self.changeList:
      # merge changes together (will be saved with changeset)
      changes = self.changeList
      accum = changes.pop(0)
      while changes:
        accum = mergeDicts(accum, changes.pop(0))

      self._resourceChanges.updateChanges(accum, self.statuses, resource)

    self.result = result
    if result.readyState:
      self.localStatus = result.readyState
    return self

  def commitChanges(self):
    """
    This can be called multiple times if the configurator yields multiple times.
    Save the changes made each time.
    """
    changes = AttributeManager.commitChanges(self)
    self.changeList.append(changes)
    return changes

  def hasInputsChanged(self):
    """
    Evaluate configuration spec's inputs and compare with the current inputs' values
    """
    # XXX this is really a "reconfiguration" operation, which can be distinct from 'configure'
    _parameters = None
    if self.lastConfigChange: # XXX this is never set
      changeset = self._manifest.loadConfigChange(self.lastConfigChange)
      _parameters = changeset.inputs
    if not _parameters:
      return not not self.inputs

    if set(self.inputs.keys()) != set(_parameters.keys()):
      return True #params were added or removed

    # XXX3 not all parameters need to be live
    # add an optional liveParameters attribute to config spec to specify which ones to check

    # compare old with new
    for name, val in self.inputs.items():
      if serializeValue(val) != _parameters[name]:
        return True
      # XXX if the value changed since the last time we checked
      #if Dependency.hasValueChanged(val, lastChecked):
      #  return True
    return False

  def hasDependenciesChanged(self):
    return any(d.hasChanged(self) for d in self.dependencies.values())

  def refreshDependencies(self):
    for d in self.dependencies.values():
      d.refresh(self)

  def summary(self):
    if self.target.name != self.target.template.name:
      rname = "%s (%s)" % (self.target.name != self.target.template.name)
    else:
      rname = self.target.name

    if self.configSpec.name != self.configSpec.className:
      cname = "%s (%s)" % (self.configSpec.name, self.configSpec.className)
    else:
      cname = self.configSpec.name
    return ("Run {action} on resource {rname} (type {rtype}, status {rstatus}) "
      + "using configurator {cname}, priority {priority}, reason: {reason}").format(
        action=self.configSpec.action,
        rname=rname,
        rtype=self.target.template.type,
        rstatus=self.target.status.name,
        cname=cname,
        priority=self.priority.name,
        reason=self.reason
      )

  def __repr__(self):
    return "ConfigTask(%s:%s %s)" % (
      self.target,
      self.configSpec.name,
      self.reason or 'unknown'
    )

class JobOptions(object):
  """
  Options available to select which tasks are run, e.g. read-only

  does the config apply to the action?
  is it out of date?
  is it in a ok state?
  """
  defaults = dict(
    parentJob=None,
    startTime=None,
    out=None,
    verbose=0,

    resource=None,
    resources=None,
    template=None,
    useConfigurator=False,

    # default options:
    add=True, # add new templates
    update=True, # run configurations that whose spec has changed but don't require a major version change
    repair="error", # or 'degraded' or "notapplied" or "none", run configurations that are not operational and/or degraded

    upgrade=False, # run configurations with major version changes or whose spec has changed
    all=False, # (re)run all configurations
    verify=False, # XXX3 discover first and set status if it differs from expected state
    readonly=False, # only run configurations that won't alter the system
    dryrun=False, # XXX2
    planOnly=False,

    requiredOnly=False,
    revertObsolete=False, #revert

    append=None,
    replace=None,
    cmdline=None
    )

  def __init__(self, **kw):
    options = self.defaults.copy()
    options.update(kw)
    self.__dict__.update(options)

class Job(ConfigChange):
  """
  runs ConfigTasks and Jobs
  """

  def __init__(self, runner, rootResource, plan, jobOptions):
    super(Job, self).__init__(Status.ok)
    assert isinstance(jobOptions, JobOptions)
    self.__dict__.update(jobOptions.__dict__)
    self.dryRun = jobOptions.dryrun
    if self.startTime is None:
      self.startTime = datetime.datetime.now()
    self.jobOptions = jobOptions
    self.runner = runner
    self.plan = plan
    self.rootResource = rootResource
    self.jobRequestQueue = []
    self.unexpectedAbort = None
    # note: tasks that never run will all share this changeid
    self.changeId = runner.incrementChangeId()
    self.parentId = self.parentJob.changeId if self.parentJob else None
    self.workDone = collections.OrderedDict()

  def createTask(self, configSpec, target, parentId=None, reason=None):
    # XXX2 if 'via'/runsOn set, create remote task instead
    task = ConfigTask(self, configSpec, target, parentId, reason = reason)
    return task

  def filterConfig(self, config, target):
    opts = self.jobOptions
    if opts.readonly and config.intent != 'discover':
      return None, "read only"
    if opts.requiredOnly and not config.required:
      return None, "required"
    if opts.resource and target.name != opts.resource:
      return None, "resource"
    if opts.resources and target.name not in opts.resources:
      return None, "resources"
    return config, None

  def getCandidateTasks(self):
    # XXX plan might call job.runJobRequest(configuratorJob) before yielding
    for (configSpec, target, reason) in self.plan.executePlan():
      configSpecName = configSpec.name
      configSpec, filterReason = self.filterConfig(configSpec, target)
      if not configSpec:
        logger.debug("skipping configspec %s for %s: doesn't match %s filter", configSpecName, target.name, filterReason)
        continue
      if self.runner.isConfigAlreadyHandled(configSpec):
        # configuration may have premptively run while executing another task
        logger.debug("configspec %s for target %s already handled", configSpecName, target.name)
        continue

      yield self.createTask(configSpec, target, reason=reason)

  def run(self):
    for task in self.getCandidateTasks():
      self.runner.addWork(task)
      if not self.shouldRunTask(task):
        continue

      if self.jobOptions.planOnly:
        if not self.cantRunTask(task):
          logger.info(task.summary())
      else:
        logger.info("Running task %s", task)
        self.runTask(task)

      if self.shouldAbort(task):
        return self.rootResource

    # the only jobs left will be those that were added to resources already iterated over
    # and were not yielding inside runTask
    while self.jobRequestQueue:
      jobRequest = self.jobRequestQueue[0]
      job = self.runJobRequest(jobRequest)
      if self.shouldAbort(job):
        return self.rootResource

    # XXX
    # if not self.parentJob:
    #   # create a job that will re-run configurations whose parameters or runtime dependencies have changed
    #   # ("config changed" tasks)
    #   # XXX3 check for orphaned resources and mark them as orphaned
    #   #  (a resource is orphaned if it was added as a dependency and no longer has dependencies)
    #   #  (orphaned resources can be deleted by the configuration that created them or manages that type)
    #   maxloops = 10 # XXX3 better loop detection
    #   for count in range(maxloops):
    #     jobOptions = JobOptions(parentJob=self, repair='none')
    #     plan = Plan(self.rootResource, self.runner.manifest.tosca, jobOptions)
    #     job = Job(self.runner, self.rootResource, plan, jobOptions)
    #     job.run()
    #     # break when there are no more tasks to run
    #     if not len(job.workDone) or self.shouldAbort(job):
    #       break
    #   else:
    #     raise UnfurlError("too many final dependency runs")

    return self.rootResource

  def runJobRequest(self, jobRequest):
    self.jobRequestQueue.remove(jobRequest)
    resourceNames=[r.name for r in jobRequest.resources]
    jobOptions = JobOptions(parentJob=self, repair='none',
      resources=resourceNames)
    plan = Plan(self.rootResource.root, self.runner.manifest.tosca, jobOptions)
    childJob = Job(self.runner, self.rootResource.root, plan, jobOptions)
    assert childJob.parentJob is self
    childJob.run()
    return childJob

  def shouldRunTask(self, task):
    """
    Checked at runtime right before each task is run
    """
    try:
      priority = task.configurator.shouldRun(task)
    except Exception:
      #unexpected error don't run this
      UnfurlTaskError(task, "shouldRun failed unexpectedly", True)
      return False

    priority = toEnum(Priority, priority, Priority.ignore)
    if priority != task.priority:
      logger.debug("configurator changed task %s priority from %s to %s", task, task.priority, priority)
      task.priority = priority
    return priority > Priority.ignore

  def cantRunTask(self, task):
    """
    Checked at runtime right before each task is run

    * validate inputs
    * check pre-conditions to see if it can be run
    * check task if it can be run
    """
    try:
      canRun = False
      reason = ''
      dependencies = list(task.target.getOperationalDependencies())
      missing = [dep for dep in dependencies if not dep.operational and dep.required]
      if missing:
        reason = "missing required dependencies: %s" % ','.join([dep.name for dep in missing])
      else:
        errors = task.configSpec.findInvalidateInputs(task.inputs)
        if errors:
          reason = 'invalid inputs: %s' % str(errors)
        else:
          preErrors = task.configSpec.findInvalidPreconditions(task.target)
          if preErrors:
            reason = 'invalid preconditions: %s' % str(preErrors)
          else:
            errors = task.configurator.cantRun(task)
            if errors:
              reason = 'configurator declined: %s' % str(errors)
            else:
              canRun = True
    except Exception:
      UnfurlTaskError(task, "cantRun failed unexpectedly", True)
      reason = 'unexpected exception in cantRun'
      canRun = False

    if canRun:
      return False
    else:
      logger.info("could not run task %s: %s", task, reason)
      return "could not run: " + reason

  def shouldAbort(self, task):
    return False #XXX3

  def summary(self):
    outputString = ''
    outputs = self.getOutputs()
    if outputs:
      outputString = ('\nOutputs:\n    ' +
      '\n    '.join("%s: %s" % (name, value) for name, value in serializeValue(outputs).items()))

    if not self.workDone:
      return 'Job %s completed: %s. Found nothing to do. %s' % (self.changeId, self.status.name, outputString)

    def format(name, task):
      required = '[required]' if task.required else ''
      return "%s: %s:%s: %s" % (name, required, task.status.name, task.result or "skipped")

    line1 = 'Job %s completed: %s. Tasks:\n    ' % (self.changeId, self.status.name)
    tasks = '\n    '.join(format(name, task) for name, task in self.workDone.items())
    return (line1 + tasks + outputString)

  def getOperationalDependencies(self):
    # XXX3 this isn't right, root job might have too many and child job might not have enough
    # plus dynamic configurations probably shouldn't be included if yielded by a configurator
    for task in self.workDone.values():
      yield task

  def getOutputs(self):
    return self.rootResource.outputs.attributes

  def runQuery(self, query, trace=0):
    from .eval import evalForFunc, RefContext
    return evalForFunc(query, RefContext(self.rootResource, trace=trace))

  def runTask(self, task):
    """
    During each task run:
    * Notification of metadata changes that reflect changes made to resources
    * Notification of add or removing dependency on a resource or properties of a resource
    * Notification of creation or deletion of a resource
    * Requests a resource with requested metadata, if it doesn't exist, a task is run to make it so
    (e.g. add a dns entry, install a package).
    """
    # XXX3 recursion or loop detection
    errors = self.cantRunTask(task)
    if errors:
      return task.finished(ConfiguratorResult(False, False, result=errors))

    task.start()
    change = None
    while True:
      try:
        result = task.send(change)
      except Exception:
        UnfurlTaskError(task, "configurator.run failed", True)
        # assume the worst
        return task.finished(ConfiguratorResult(True, True, Status.error))
      if isinstance(result, TaskRequest):
        subtask = self.createTask(result.configSpec, result.target, self.changeId)
        self.runner.addWork(subtask)
        change = self.runTask(subtask) # returns a configuration
      elif isinstance(result, JobRequest):
        job = self.runJobRequest(result)
        change = job
      elif isinstance(result, ConfiguratorResult):
        retVal = task.finished(result)
        logger.info("finished running task %s: %s; %s", task, task.target.status, result)
        return retVal
      else:
        UnfurlTaskError(task, 'unexpected result from configurator', True)
        return task.finished(ConfiguratorResult(True, True, Status.error))

class Runner(object):
  def __init__(self, manifest):
    self.manifest = manifest
    assert self.manifest.tosca
    self.lastChangeId = manifest.lastChangeId
    self.currentJob = None

  def addWork(self, task):
    key = "%s:%s:%s:%s" % (task.target.name, task.configSpec.name, task.configSpec.action, task.changeId)
    self.currentJob.workDone[key] = task
    task.job.workDone[key] = task

  def isConfigAlreadyHandled(self, configSpec):
    return False # XXX
    # return configSpec.name in self.currentJob.workDone

  def createJob(self, joboptions):
    """
    Selects task to run based on job options and starting state of manifest
    """
    root = self.manifest.getRootResource()
    assert self.manifest.tosca
    plan = Plan(root, self.manifest.tosca, joboptions)
    return Job(self, root, plan, joboptions)

  def incrementChangeId(self):
    self.lastChangeId +=1
    return self.lastChangeId

  def run(self, jobOptions=None):
    """
    """
    if jobOptions is None:
      jobOptions = JobOptions()
    job = self.createJob(jobOptions)
    self.currentJob = job
    try:
      ansibleDisplay.verbosity = jobOptions.verbose
      job.run()
    except Exception:
      job.localStatus = Status.error
      job.unexpectedAbort = UnfurlError("unexpected exception while running job", True, True)
    self.currentJob = None
    self.manifest.commitJob(job)
    return job



"""
This module defines the core model and implements the runtime operations of the model.

The state of the system is represented as a collection of Resources.
Each resource have a status; attributes that describe its state; and a TOSCA template
 which describes its capabilities, relationships and available interfaces for configuring and interacting with it.
"""
import six
import collections
from itertools import chain

from ansible.template import Templar
from ansible.parsing.dataloader import DataLoader

from .util import (UnfurlError, loadClass, toEnum)
from .result import ResourceRef, ChangeAware
#from .local import LocalEnv
from .support import AttributeManager, Defaults, Status, Action, Priority
from .tosca import CapabilitySpec, RelationshipSpec, NodeSpec, TopologySpec

import logging
logger = logging.getLogger('unfurl')

# CF 3.4.1 Node States p61

S = Status
A = Action

class Operational(ChangeAware):
  """
  This is an abstract base class for Jobs, Resources, and Configurations all have a Status associated with them
  and all use the same algorithm to compute their status from their dependent resouces, tasks, and configurations

  # operational: boolean: ok or degraded
  # status: ok, degraded, error, notpresent
  # degraded: non-fatal errors or didn't provide required attributes or if couldnt upgrade
  """

  # XXX3 rename Status to ReadyState, add setter on status instead of setting localStatus?
  # XXX3 add repairable, messages?

  # core properties to override
  @property
  def priority(self):
    return Defaults.shouldRun

  @property
  def localStatus(self):
    return Status.notapplied

  def getOperationalDependencies(self):
    return ()

  @property
  def manualOverideStatus(self):
    return None

  # derived properties:
  @property
  def operational(self):
    return self.status == Status.ok or self.status == Status.degraded

  @property
  def active(self):
    return self.status <= Status.error

  @property
  def status(self):
    if self.manualOverideStatus is not None:
      status = self.manualOverideStatus
      if status >= Status.error:
        return status
    else:
      status = self.localStatus
      if status == Status.error or status == Status.notpresent:
        return status

    return self.aggregateStatus(self.getOperationalDependencies(), status)

  @property
  def required(self):
    return self.priority == Priority.required

  @property
  def lastStateChange(self):
    return None

  @property
  def lastConfigChange(self):
    return None

  @property
  def lastChange(self):
    if not self.lastStateChange:
      return self.lastConfigChange
    elif not self.lastConfigChange:
      return self.lastStateChange
    else:
      return max(self.lastStateChange, self.lastConfigChange)

  def hasChanged(self, changeset):
    # if changed since the last time we checked
    if not self.lastChange:
      return False
    if not changeset:
      return True
    return self.lastChange > changeset.changeId

  @staticmethod
  def aggregateStatus(statuses, defaultStatus = Status.notapplied):
    # error if a configuration is required and not operational
    # error if a non-configuration managed child resource is required and not operational
    # degraded if non-required configurations and resources are not operational
    #          or required configurations and resources are degraded
    # ok otherwise

    #any required is pending then aggregate is pending
    #any other are pending only set pending if aggregate is ok
    aggregate = defaultStatus if defaultStatus else Status.notapplied
    for status in statuses:
      assert isinstance(status, Operational), status
      if status.priority == Priority.ignore:
        continue
      elif status.required:
        if not status.operational:
          aggregate = Status.error
          break
        elif status.status == Status.degraded:
          aggregate = Status.degraded
      elif status.status == Status.pending and aggregate <= Status.pending:
        aggregate = Status.pending
      elif not status.operational:
        aggregate = Status.degraded
      elif aggregate == Status.notapplied:
        aggregate = Status.ok
    return aggregate

class OperationalInstance(Operational):
  def __init__(self, status=None, priority=None, manualOveride=None, lastStateChange=None, lastConfigChange=None):
    if isinstance(status, OperationalInstance):
      self._localStatus = status._localStatus
      self._manualOverideStatus = status._manualOverideStatus
      self._priority = status._priority
      self._lastStateChange = status._lastStateChange
      self._lastConfigChange = status._lastConfigChange
      self.dependencies = status.dependencies
    else:
      self._localStatus = toEnum(Status, status)
      self._manualOverideStatus = toEnum(Status, manualOveride)
      self._priority = toEnum(Priority, priority)
      self._lastStateChange = lastStateChange
      self._lastConfigChange = lastConfigChange
      self.dependencies = []
    #self.repairable = False # XXX3
    #self.messages = [] # XXX3

  def getOperationalDependencies(self):
    return self.dependencies

  def localStatus():
    doc = "The localStatus property."
    def fget(self):
      return self._localStatus
    def fset(self, value):
      self._localStatus = value
    def fdel(self):
      del self._localStatus
    return locals()
  localStatus = property(**localStatus())

  def manualOverideStatus():
    doc = "The manualOverideStatus property."
    def fget(self):
      return self._manualOverideStatus
    def fset(self, value):
      self._manualOverideStatus = value
    def fdel(self):
      del self._manualOverideStatus
    return locals()
  manualOverideStatus = property(**manualOverideStatus())

  def priority():
    doc = "The priority property."
    def fget(self):
      return Defaults.shouldRun if self._priority is None else self._priority
    def fset(self, value):
      self._priority = value
    def fdel(self):
      del self._priority
    return locals()
  priority = property(**priority())

  @property
  def lastStateChange(self):
    return self._lastStateChange

  @property
  def lastConfigChange(self):
    return self._lastConfigChange

class _ChildResources(collections.Mapping):
  def __init__(self, resource):
    self.resource = resource

  def __getitem__(self, key):
    return self.resource.findResource(key)

  def __iter__(self):
    return self.resource.getSelfAndDescendents()

  def __len__(self):
    return len(tuple(self))

class NodeInstance(OperationalInstance, ResourceRef):
  attributeManager = None

  def __init__(self, name='', attributes=None, parent=None, template=None, status=None):
    OperationalInstance.__init__(self, status)
    self.name = name
    self._attributes = attributes or {}
    self.parent = parent
    if parent:
      getattr(parent, self.parentRelation).append(self)

    self.template = template or self.templateType()

  def _resolve(self, key):
    # might return a Result
    self.attributes[key] # force resolve
    return self.attributes._attributes[key]

  def localStatus():
    doc = "The localStatus property."
    def fget(self):
      return self._localStatus
    def fset(self, value):
      if self.root.attributeManager:
        self.root.attributeManager.setStatus(self, value)
      self._localStatus = value
    def fdel(self):
      del self._localStatus
    return locals()
  localStatus = property(**localStatus())

  @property
  def key(self):
    return  "%s::.%s::%s" % (self.parent.key, self.parentRelation, self.name)

  def asRef(self, options=None):
    return {"ref": self.key}

  @property
  def attributes(self):
    """
    attributes should be live values but _attributes will be serialized value
    """
    if not self.root.attributeManager:
      if not self.attributeManager:
        # inefficient but create a local one for now
        self.attributeManager = AttributeManager()
      # XXX3 changes to self.attributes aren't saved
      return self.attributeManager.getAttributes(self)

    # returned registered attribute or create a new one
    # attribute class getter resolves references
    return self.root.attributeManager.getAttributes(self)

# both have occurrences
# only need to configure capabilities as required by a relationship
class Capability(NodeInstance):
  # 3.7.2 Capability definition p. 97
  # 3.8.1 Capability assignment p. 114
  parentRelation = 'capabilities'
  templateType = CapabilitySpec

class Relationship(NodeInstance):
  # 3.7.3 Requirements definition p. 99
  # 3.8.2 Requirements assignment p. 115
  parentRelation = 'requirements'
  templateType = RelationshipSpec
  target = None #XXX

  # XXX add a target attribute
  def getOperationalDependencies(self):
    if self.target:
      yield self.target

class Resource(NodeInstance):
  """
      capabilities:
      name:
      attributes
      dependencies: #includes requirements
  """
  baseDir = ''
  createdOn = None
  templateType = NodeSpec
  parentRelation = 'resources'
  # createdFrom = None

  # spec is a NodeTemplate
  def __init__(self, name='', attributes=None, parent=None, template=None, status=None):
    if parent:
      if parent.root.findResource(name):
        raise UnfurlError('can not create resource name "%s" is already in use' % name)
    self.resources = []
    self.capabilities = []
    self.requirements = []
    NodeInstance.__init__(self, name, attributes, parent, template, status)

    if self.root is self:
      self._all = _ChildResources(self)
      self._templar = Templar(DataLoader())

    self._interfaces = {} #XXX2 exclude when pickling
    # preload
    self.getInterface('inherit')
    self.getInterface('default')

  def _resolve(self, key):
    # might return a Result
    try:
      self.attributes[key] # force resolve
      return self.attributes._attributes[key]
    except KeyError:
      try:
        inherit = self._interfaces.get('inherit') # pre-loaded
        if inherit:
          return inherit(key)
        else:
          raise
      except KeyError:
        default = self._interfaces.get('default') # pre-loaded
        if default:
          return default(key)
        else:
          raise

  @property
  def key(self):
    return  "::%s"% self.name

  def getOperationalDependencies(self):
    # XXX
    #for instance in chain(self.capabilities, self.requirements):
    #  yield instance
    if self.parent and self.parent is not self.root:
      yield self.parent

  def getSelfAndDescendents(self):
    "Recursive descendent including self"
    yield self
    for r in self.resources:
      for descendent in r.getSelfAndDescendents():
        yield descendent

  @property
  def descendents(self):
    return list(self.getSelfAndDescendents())

  def findResource(self, resourceid):
    if self.name == resourceid:
      return self
    for r in self.resources:
      child = r.findResource(resourceid)
      if child:
        return child
    return None

  def setBaseDir(self, baseDir):
    self.baseDir = baseDir
    if not self._templar or self._templar._basedir != baseDir:
      loader = DataLoader()
      loader.set_basedir(baseDir)
      self._templar = Templar(loader)

  def addInterface(self, klass, name=None):
    if not isinstance(klass, six.string_types):
      klass = klass.__module__ + '.' + klass.__name__
    current = self._attributes.setdefault('.interfaces', {})
    current[name or klass] = klass
    return current

  def getInterface(self, name):
    # XXX uses TOSCA interfaces instead
    if '.interfaces' not in self._attributes:
      return None # no interfaces

    if not isinstance(name, six.string_types):
      name = name.__module__ + '.' + name.__name__
    instance = self._interfaces.get(name)
    if instance:
      return instance
    else:
      className = self._attributes['.interfaces'].get(name)
      if className:
        instance = loadClass(className)(name, self)
        self._interfaces[name] = instance
        return instance
    return None

  def __eq__(self, other):
    if self is other:
      return True
    if not isinstance(other, Resource):
      return False
    if not self.lastChange:
      # only support equality if resource has a changeid
      return False
    return self.name == other.name and self.lastChange == other.lastChange

  def __repr__(self):
    return "Resource('%s')" % self.name

class TopologyResource(Resource):
  templateType = TopologySpec

  def __init__(self, template, status=None):
    Resource.__init__(self, 'root', template=template, status=status)
    # add these as special child resources so they can be accessed like "::inputs::foo"
    self.inputs = Resource('inputs', template.inputs, self)
    self.outputs = Resource('outputs', template.outputs, self)

  def getOperationalDependencies(self):
    for instance in self.resources:
      if instance.name not in ['inputs', 'outputs']:
        yield instance



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

setup(
    setup_requires=['pbr'],
    pbr=True,
    # XXX don't include .py not committed to git
    include_package_data=True
)



import unittest
from unfurl.yamlmanifest import YamlManifest
from unfurl.util import UnfurlError, UnfurlValidationError

class ManifestSyntaxTest(unittest.TestCase):
  def test_hasVersion(self):
    hasVersion = """
    apiVersion: unfurl/v1alpha1
    kind: Manifest
    spec: {}
    """
    assert YamlManifest(hasVersion)

  def test_validateVersion(self):
    badVersion = """
    apiVersion: 2
    kind: Manifest
    spec: {}
    """
    with self.assertRaises(UnfurlError) as err:
      YamlManifest(badVersion)
    self.assertIn("2 is not one of [\'unfurl/v1alpha1\']", str(err.exception))

    missingVersion = """
    spec: {}
    """
    with self.assertRaises(UnfurlError) as err:
      YamlManifest(missingVersion)
    self.assertIn("'apiVersion' is a required property", str(err.exception)) #, <ValidationError: "'kind' is a required property">]''')

  @unittest.skip("TODO")
  def test_validResourceNames(self):
    #should only contain [a-z0-9]([a-z0-9\-]*[a-z0-9])?
    pass

  def test_template_inheritance(self):
    manifest = '''
apiVersion: unfurl/v1alpha1
kind: Manifest
templates:
  base:
    configurations:
      step1: {}
spec:
    +templates/production:
'''
    with self.assertRaises(UnfurlError) as err:
      YamlManifest(manifest)
    self.assertIn('missing includes: [templates/production]', str(err.exception))

    manifest = '''
apiVersion: unfurl/v1alpha1
kind: Manifest

configurators:
  step1:
    actions:
      install: foo
  step2:
    actions:
      install: foo
templates:
  base:
    configurations:
      step1:
        className: foo
        majorVersion: 0
root:
 resources:
  cloud3:
    spec:
      +templates/base:
      # overrides and additions from templates
      configurations:
        step1: {}
'''
    # XXX update yaml
    #overrides base.step1 defination, doesn't add a component
    # manifestObj = YamlManifest(manifest)
    # assert len(manifestObj.rootResource.all['cloud3'].spec['configurations']) == 1, manifestObj.rootResource.all['cloud3'].spec['configurations']

  @unittest.skip("update syntax")
  def test_override(self):
    #component names have to be qualified to override
    #duplicate names both run with distinct values
    manifest = '''
apiVersion: unfurl/v1alpha1
kind: Manifest

configurators:
  step2:
    actions:
      install: foo
templates:
  base:
    templates:
    configurations:
      step1:
        className: foo
        majorVersion: 0
        configurator:
          parameterSchema:
            - name: test
              default: default
        parameters:
          test:
            base
root:
 resources:
  cloud3:
    spec:
      templates:
        - base
      # overrides and additions from template
      configurations:
        +templates/base/configurations:
        step1:
          parameters:
            test: derived
'''
    configurations = YamlManifest(manifest).rootResource.all['cloud3'].spec['configurations']
    self.assertEqual(list(configurations.values())[0].parameters, {
      'test': 'derived',
      # '+%': 'replaceProps'
    })

  @unittest.skip("update syntax")
  def test_missingConfigurator(self):
    manifest = '''
apiVersion: unfurl/v1alpha1
kind: Manifest

configurators:
  step2:
    actions:
      install: foo
templates:
  base:
    configurations:
      - step1
      - step2
root:
 resources:
  cloud3:
    spec:
      templates:
        - base
      configurations:
        +templates/base/configurations:

'''
    with self.assertRaises(UnfurlValidationError) as err:
      YamlManifest(manifest)
    self.assertIn(str(err.exception.errors), '''[<ValidationError: "['step1', 'step2'] is not of type 'object'">]''')

#   def test_badparams(self):
#     # don't match spec definition
#     manifest = '''
# apiVersion: unfurl/v1alpha1
# kind: Manifest
#
# configurators:
#   step2:
#     actions:
#       install: foo
# templates:
#   base:
#     configurations:
#       - name: step1
#         configurator:
#           parameterSchema:
#             - name: test
#               type: string
#               default: default
#         parameters:
#           test: base
# resources:
#   cloud3:
#     spec:
#       templates:
#         - base
#       # overrides and additions
#       configurations:
#         - name: base.step1
#           parameters:
#             # error: should be a string
#             test: 0
# '''
#     with self.assertRaises(UnfurlValidationError) as err:
#       m = YamlManifest(manifest)
#     self.assertEqual(err and str(err.exception.errors[0][0]), "invalid value")
#
#   def test_unexpectedParam(self):
#     #parameter missing from spec
#     manifest = '''
# apiVersion: unfurl/v1alpha1
# kind: Manifest
#
# configurators:
#   step2:
#     actions:
#       install: foo
# templates:
#   base:
#     configurations:
#       - name: step1
#         configurator:
#           parameterSchema:
#             - name: test
#               default: default
#         parameters:
#           test: base
# resources:
#   cloud3:
#     spec:
#       templates:
#         - base
#       # overrides and additions
#       configurations:
#         - name: base.step1
#           parameters:
#             doesntexist: True
# '''
#     with self.assertRaises(UnfurlValidationError) as err:
#       YamlManifest(manifest)
#     self.assertEqual(str(err.exception.errors[0][0]), "unexpected parameters")
#
#   def test_missingParam(self):
#     #missing required parameter
#     manifest = '''
# apiVersion: unfurl/v1alpha1
# kind: Manifest
#
# configurators:
#   step2:
#     actions:
#       install: foo
# templates:
#   base:
#     configurations:
#       - name: step1
#         configurator:
#           parameterSchema:
#             - name: test
#               required: True
# resources:
#   cloud3:
#     spec:
#       templates:
#         - base
#       # overrides and additions
#       configurations:
#         - name: base.step1
# '''
#     with self.assertRaises(UnfurlValidationError) as err:
#       YamlManifest(manifest)
#     self.assertEqual(str(err.exception.errors[0][0]), "missing required parameter")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os.path
import requests
import shutil
import six
import tempfile
import yaml
import zipfile

from toscaparser.common.exception import ExceptionCollector
from toscaparser.common.exception import URLException
from toscaparser.common.exception import ValidationError
from toscaparser.imports import ImportsLoader
from toscaparser.utils.gettextutils import _
from toscaparser.utils.urlutils import UrlUtils

try:  # Python 2.x
    from BytesIO import BytesIO
except ImportError:  # Python 3.x
    from io import BytesIO


class CSAR(object):

    def __init__(self, csar_file, a_file=True):
        self.path = csar_file
        self.a_file = a_file
        self.is_validated = False
        self.error_caught = False
        self.csar = None
        self.temp_dir = None

    def validate(self):
        """Validate the provided CSAR file."""

        self.is_validated = True

        # validate that the file or URL exists
        missing_err_msg = (_('"%s" does not exist.') % self.path)
        if self.a_file:
            if not os.path.isfile(self.path):
                ExceptionCollector.appendException(
                    ValidationError(message=missing_err_msg))
                return False
            else:
                self.csar = self.path
        else:  # a URL
            if not UrlUtils.validate_url(self.path):
                ExceptionCollector.appendException(
                    ValidationError(message=missing_err_msg))
                return False
            else:
                response = requests.get(self.path)
                self.csar = BytesIO(response.content)

        # validate that it is a valid zip file
        if not zipfile.is_zipfile(self.csar):
            err_msg = (_('"%s" is not a valid zip file.') % self.path)
            ExceptionCollector.appendException(
                ValidationError(message=err_msg))
            return False

        # validate that it contains the metadata file in the correct location
        self.zfile = zipfile.ZipFile(self.csar, 'r')
        filelist = self.zfile.namelist()
        if 'TOSCA-Metadata/TOSCA.meta' not in filelist:
            err_msg = (_('"%s" is not a valid CSAR as it does not contain the '
                         'required file "TOSCA.meta" in the folder '
                         '"TOSCA-Metadata".') % self.path)
            ExceptionCollector.appendException(
                ValidationError(message=err_msg))
            return False

        # validate that 'Entry-Definitions' property exists in TOSCA.meta
        data = self.zfile.read('TOSCA-Metadata/TOSCA.meta')
        invalid_yaml_err_msg = (_('The file "TOSCA-Metadata/TOSCA.meta" in '
                                  'the CSAR "%s" does not contain valid YAML '
                                  'content.') % self.path)
        try:
            meta = yaml.load(data)
            if type(meta) is dict:
                self.metadata = meta
            else:
                ExceptionCollector.appendException(
                    ValidationError(message=invalid_yaml_err_msg))
                return False
        except yaml.YAMLError:
            ExceptionCollector.appendException(
                ValidationError(message=invalid_yaml_err_msg))
            return False

        if 'Entry-Definitions' not in self.metadata:
            err_msg = (_('The CSAR "%s" is missing the required metadata '
                         '"Entry-Definitions" in '
                         '"TOSCA-Metadata/TOSCA.meta".')
                       % self.path)
            ExceptionCollector.appendException(
                ValidationError(message=err_msg))
            return False

        # validate that 'Entry-Definitions' metadata value points to an
        # existing file in the CSAR
        entry = self.metadata.get('Entry-Definitions')
        if entry and entry not in filelist:
            err_msg = (_('The "Entry-Definitions" file defined in the '
                         'CSAR "%s" does not exist.') % self.path)
            ExceptionCollector.appendException(
                ValidationError(message=err_msg))
            return False

        # validate that external references in the main template actually
        # exist and are accessible
        self._validate_external_references()
        return not self.error_caught

    def get_metadata(self):
        """Return the metadata dictionary."""

        # validate the csar if not already validated
        if not self.is_validated:
            self.validate()

        # return a copy to avoid changes overwrite the original
        return dict(self.metadata) if self.metadata else None

    def _get_metadata(self, key):
        if not self.is_validated:
            self.validate()
        return self.metadata.get(key)

    def get_author(self):
        return self._get_metadata('Created-By')

    def get_version(self):
        return self._get_metadata('CSAR-Version')

    def get_main_template(self):
        entry_def = self._get_metadata('Entry-Definitions')
        if entry_def in self.zfile.namelist():
            return entry_def

    def get_main_template_yaml(self):
        main_template = self.get_main_template()
        if main_template:
            data = self.zfile.read(main_template)
            invalid_tosca_yaml_err_msg = (
                _('The file "%(template)s" in the CSAR "%(csar)s" does not '
                  'contain valid TOSCA YAML content.') %
                {'template': main_template, 'csar': self.path})
            try:
                tosca_yaml = yaml.load(data)
                if type(tosca_yaml) is not dict:
                    ExceptionCollector.appendException(
                        ValidationError(message=invalid_tosca_yaml_err_msg))
                return tosca_yaml
            except Exception:
                ExceptionCollector.appendException(
                    ValidationError(message=invalid_tosca_yaml_err_msg))

    def get_description(self):
        desc = self._get_metadata('Description')
        if desc is not None:
            return desc

        self.metadata['Description'] = \
            self.get_main_template_yaml().get('description')
        return self.metadata['Description']

    def decompress(self):
        if not self.is_validated:
            self.validate()
        self.temp_dir = tempfile.NamedTemporaryFile().name
        with zipfile.ZipFile(self.csar, "r") as zf:
            zf.extractall(self.temp_dir)

    def _validate_external_references(self):
        """Extracts files referenced in the main template

        These references are currently supported:
        * imports
        * interface implementations
        * artifacts
        """
        try:
            self.decompress()
            main_tpl_file = self.get_main_template()
            if not main_tpl_file:
                return
            main_tpl = self.get_main_template_yaml()

            if 'imports' in main_tpl:
                ImportsLoader(main_tpl['imports'],
                              os.path.join(self.temp_dir, main_tpl_file))

            if 'topology_template' in main_tpl:
                topology_template = main_tpl['topology_template']

                if 'node_templates' in topology_template:
                    node_templates = topology_template['node_templates']

                    for node_template_key in node_templates:
                        node_template = node_templates[node_template_key]
                        if 'artifacts' in node_template:
                            artifacts = node_template['artifacts']
                            for artifact_key in artifacts:
                                artifact = artifacts[artifact_key]
                                if isinstance(artifact, six.string_types):
                                    self._validate_external_reference(
                                        main_tpl_file,
                                        artifact)
                                elif isinstance(artifact, dict):
                                    if 'file' in artifact:
                                        self._validate_external_reference(
                                            main_tpl_file,
                                            artifact['file'])
                                else:
                                    ExceptionCollector.appendException(
                                        ValueError(_('Unexpected artifact '
                                                     'definition for "%s".')
                                                   % artifact_key))
                                    self.error_caught = True
                        if 'interfaces' in node_template:
                            interfaces = node_template['interfaces']
                            for interface_key in interfaces:
                                interface = interfaces[interface_key]
                                for opertation_key in interface:
                                    operation = interface[opertation_key]
                                    if isinstance(operation, six.string_types):
                                        self._validate_external_reference(
                                            main_tpl_file,
                                            operation,
                                            False)
                                    elif isinstance(operation, dict):
                                        if 'implementation' in operation:
                                            self._validate_external_reference(
                                                main_tpl_file,
                                                operation['implementation'])
        finally:
            if self.temp_dir:
                shutil.rmtree(self.temp_dir)

    def _validate_external_reference(self, tpl_file, resource_file,
                                     raise_exc=True):
        """Verify that the external resource exists

        If resource_file is a URL verify that the URL is valid.
        If resource_file is a relative path verify that the path is valid
        considering base folder (self.temp_dir) and tpl_file.
        Note that in a CSAR resource_file cannot be an absolute path.
        """
        if UrlUtils.validate_url(resource_file):
            msg = (_('The resource at "%s" cannot be accessed.') %
                   resource_file)
            try:
                if UrlUtils.url_accessible(resource_file):
                    return
                else:
                    ExceptionCollector.appendException(
                        URLException(what=msg))
                    self.error_caught = True
            except Exception:
                ExceptionCollector.appendException(
                    URLException(what=msg))
                self.error_caught = True

        if os.path.isfile(os.path.join(self.temp_dir,
                                       os.path.dirname(tpl_file),
                                       resource_file)):
            return

        if raise_exc:
            ExceptionCollector.appendException(
                ValueError(_('The resource "%s" does not exist.')
                           % resource_file))
            self.error_caught = True



import unfurl.util
import unittest
from unfurl.yamlmanifest import YamlManifest
from unfurl.job import Runner, JobOptions
from unfurl.ansibleconfigurator import runPlaybooks
from unfurl.runtime import Status
import os
import os.path
import warnings
import sys

class AnsibleTest(unittest.TestCase):
  def setUp(self):
    # need to call this again on python 2.7:
    unfurl.util.initializeAnsible()
    try:
      # Ansible generates tons of ResourceWarnings
      warnings.simplefilter("ignore", ResourceWarning)
    except:
      # python 2.x doesn't have ResourceWarning
      pass
    self.results = {}

  def runPlaybook(self, args=None):
    return runPlaybooks([os.path.join(os.path.dirname(__file__),'testplaybook.yaml')], 'localhost,', {
    'ansible_connection': 'local',
    'extra': 1
    }, args)

  def test_runplaybook(self):
    results = self.runPlaybook()
    self.results['runplaybook'] = results
    self.assertEqual('test',
      results.variableManager._nonpersistent_fact_cache['localhost'].get('one_fact'))

    hostfacts = results.variableManager._nonpersistent_fact_cache['localhost']
    self.assertEqual(hostfacts['one_fact'], 'test')
    self.assertEqual(hostfacts['echoresults']['stdout'], "hello")

  def test_verbosity(self):
    # setting UNFURL_LOGGING can break this test, so skip if it's set
    if not os.getenv('UNFURL_LOGGING'):
      results = self.runPlaybook()
      # task test-verbosity was skipped
      assert not results.resultsByStatus.ok.get('test-verbosity')
      assert results.resultsByStatus.skipped.get('test-verbosity')
      results = self.runPlaybook(['-vv'])
      # task test-verbosity was ok this time
      assert results.resultsByStatus.ok.get('test-verbosity')
      assert not results.resultsByStatus.skipped.get('test-verbosity')

manifest = '''
apiVersion: unfurl/v1alpha1
kind: Manifest
configurations:
  create:
    implementation: unfurl.ansibleconfigurator.AnsibleConfigurator
    inputs:
      playbook:
        q:
          - name: Hello
            command: echo "{{hostvars['localhost'].ansible_python_interpreter}}"
spec:
  node_templates:
    test1:
      type: tosca.nodes.Root
      interfaces:
        Standard:
          +configurations:
'''

class AnsibleConfiguratorTest(unittest.TestCase):
  def setUp(self):
    try:
      # Ansible generates tons of ResourceWarnings
      warnings.simplefilter("ignore", ResourceWarning)
    except:
      # python 2.x doesn't have ResourceWarning
      pass

  def test_ansible(self):
    """
    test that runner figures out the proper tasks to run
    """
    runner = Runner(YamlManifest(manifest))
    run1 = runner.run(JobOptions(resource='test1'))
    assert not run1.unexpectedAbort, run1.unexpectedAbort.getStackTrace()
    assert len(run1.workDone) == 1, run1.workDone
    result = list(run1.workDone.values())[0].result.result
    self.assertEqual(result, {'returncode': 0, 'stdout': sys.executable})
    assert run1.status == Status.ok, run1.summary()
